/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#ifndef _DMA_H
#define _DMA_H

#include <scif.h>

#include <sys/time.h>

#include "../common/COITypes_common.h"
#include "../common/COIResult_common.h"
#include "../source/COIBuffer_source.h"
#include "../common/COIPerf_common.h"
#include "../internal/LockfreeQueue.h"
#include "../internal/_PthreadMutexAutoLock.h"
#include "../internal/_Message.h"

//Macros to define Threshold size that determines when to Use CPU instead of
//DMA for COI_COPY_UNSPECIFIED type.

//TODO: KNC Shady Cove numbers were used below. Update to optimize for diff
//      platform. The current comparisons use <= the thresholds below.
//      The thresholds are in bytes.
#define THRESHOLD_SCIF_WRITETO      63
#define THRESHOLD_SCIF_READFROM     32
#define THRESHOLD_SCIF_VWRITETO     63
#define THRESHOLD_SCIF_VREADFROM    32


// Each COIProcess has an associated fence object which is used to wait
// for DMAs to and from that process.
class COIDMAFence
{
public:
    COIDMAFence(scif_epd_t endpoint, void *process);
   ~COIDMAFence();

    COIRESULT WaitForDMA(off_t length);
    COIRESULT StoreAsyncDMA(off_t     length,
                            void*     task_node,
                            uint32_t  frag_count = 0,
                            bool      memcpy = false);

    static const uint64_t   MAXSPINSIZE = 2*1024*1024;
    static const double     MAXSPINTIME = 0.0005;

    // Wrap the calls we make to scif_fence* APIs.
    // Currently only does local fencing stuff.
    // Putting this as static in the header for now to avoid
    // more complex compilation stuff.
    // A better solution would be to put this in the
    // mechanism/message/(somewhere) , but the interface
    // to reserve a fence slot becomes difficult to expose
    // in a general way, so I'm leaving it here.
    static COIRESULT ScifFenceLocalWrapper(scif_epd_t endpoint,
                                           uint64_t length,
                                           volatile uint64_t* signal_addr,
                                           off_t signal_local_offset)
    {
        // Default to using the slower mark+wait rather than signal.
        bool use_mark_wait = true;
        int status = -1;
        UNUSED_ATTR const char* failed_scif_call = NULL;

        // If we have a signal address that we can spin on and the
        // transfer is small enough we can get faster transfers
        // by spinning in a while loop.
        if( (length <= COIDMAFence::MAXSPINSIZE) &&
                signal_addr )
        {
            use_mark_wait = false;
            *signal_addr = COIDMAFence::SLOT_RESERVED;

            status = scif_fence_signal(endpoint,
                    signal_local_offset, COIDMAFence::SLOT_SIGNALED,
                    0, 0,
                    SCIF_FENCE_INIT_SELF | SCIF_SIGNAL_LOCAL);
            if( status == -1 )
            {
                failed_scif_call = "scif_fence_signal";
                goto end;
            }
            // We requested that the value SLOT_SIGNALED be written into the signal offset,
            // which is really the same as the signal page. Let's do a while()
            // to check if the DMA finished, but also have a timeout to not
            // burn too many cycles.
            struct timeval t0, t1;
            uint64_t max_wait = COIDMAFence::MAXSPINTIME * 1000000;
            gettimeofday(&t0, NULL);
            uint64_t elapsed = 0;

            while (*signal_addr)
            {
                gettimeofday(&t1, NULL);
                elapsed = (t1.tv_sec - t0.tv_sec) * 1000000 + (t1.tv_usec - t0.tv_usec);
                if ( elapsed > max_wait )
                {
                    use_mark_wait = true;
                    break;
                }
            }
        }

        // The transfer was too big to begin with or we already spent
        // too much time in a busy while loop. Let's use the
        // scif_fence mark+wait() APIs.
        if ( use_mark_wait )
        {
            int mark;
            status = scif_fence_mark(endpoint, SCIF_FENCE_INIT_SELF, &mark);
            if( status == -1 )
            {
                failed_scif_call = "scif_fence_mark";
                goto end;
            }
            status = scif_fence_wait(endpoint,mark);
            if( status == -1 )
            {
                failed_scif_call = "scif_fence_wait";
                goto end;
            }
        }

end:
        if(status == -1)
        {
            // Don't print anything if the process died, we have a good way
            // to report that back to users.
            if( IsErrnoCOI_PROCESS_DIED(errno) )
            {
                return COI_PROCESS_DIED;
            }

            // Log the error
#ifdef COILOG_ERROR
#define LOG_ERROR COILOG_ERROR
#else
#ifdef DPRINTF
#define LOG_ERROR DPRINTF
#else
#define LOG_ERROR(...)
#endif
#endif
            LOG_ERROR("Error while signaling DMA completion."
                         "%s returned errno: %s\n", failed_scif_call, strerror(errno));
#undef LOG_ERROR

            return COI_ERROR;
        }
        return COI_SUCCESS;
    }

    //Node that gets enqueued to later on fetch related information
    //when a DMA operation finishes
    typedef struct
    {
        COIDMAFence*    fence;
        off_t     length;
        void*           task_node;
        COIEVENT        event;    // Event of the task_node.Maintain this for
                                  // for verification purposes
        bool            memcpy;   // If the copy is local to local
    } AsyncNode;


    bool SetupSignalPage();       //needs to be public so we can presetup and register memory
private:
    bool ReserveFenceSlot(off_t* offset, uint64_t** addr);
    void ReturnFenceSlot(uint64_t* addr);

    // Note:
    // Code depends on SLOT_SIGNALED = 0 and the others != 0
    enum
    {
        SLOT_SIGNALED = 0,
        SLOT_RESERVED = 1,
        SLOT_FREE     = 2
    };
    scif_epd_t      m_endpoint;
    pthread_mutex_t m_lock;
    uint64_t*       m_mem;
    off_t           m_offset;

// These are used to process asynchronous DMA requests
// Creates a thread to wait on Fence on a separate thread
// which makes the DMA operation truely asynchronous

    pthread_cond_t      m_asyncCond;
    pthread_mutex_t     m_asyncMux;
    pthread_t           m_asyncThread;
    void*               AsyncWait(void);
    LockfreeQueue<AsyncNode*>    m_asyncQueue;
    LockfreeQueue<AsyncNode*>    m_freeNodes;
    volatile bool       m_beingDestroyed;

public:
     void*              m_process; //process associated with this fence
    static void * ThreadProc(void *lpThreadParameter) {
        COIDMAFence *dma = (COIDMAFence*) lpThreadParameter;
        dma->AsyncWait();
        pthread_exit(NULL);
        return 0;
    }

};

namespace COIDMAManager
{

    COIRESULT CopyLocalToLocal( COIDMAFence*  fence,
                                void*         task_node,
                                bool          async,
                                uint32_t      frag_count);
    COIRESULT CopyToLocal(  scif_epd_t    endpoint,
                            COIDMAFence*  fence,
                            uint64_t      dst_handle,
                            off_t         dst_offset,
                            uint64_t      src_handle,
                            off_t         src_offset,
                            off_t         length,
                            COI_COPY_TYPE type = COI_COPY_UNSPECIFIED,
                            void*         task_node = NULL,
                            bool          async=false,
                            uint32_t      frag_count = 0);


    COIRESULT CopyToRemote( scif_epd_t    endpoint,
                            COIDMAFence*  fence,
                            uint64_t      dst_handle,
                            off_t         dst_offset,
                            uint64_t      src_handle,
                            off_t         src_offset,
                            off_t         length,
                            COI_COPY_TYPE type = COI_COPY_UNSPECIFIED,
                            void*         task_node = NULL,
                            bool          async=false,
                            uint32_t      frag_count = 0);

    COIRESULT VCopyToLocal( scif_epd_t    endpoint,
                            COIDMAFence*  fence,
                            void*         dst_address,
                            off_t         dst_offset,
                            uint64_t      src_handle,
                            off_t         src_offset,
                            off_t         length,
                            COI_COPY_TYPE type = COI_COPY_UNSPECIFIED,
                            void*         task_node = NULL,
                            bool          async=false,
                            uint32_t      frag_count = 0);

    COIRESULT VCopyToRemote(scif_epd_t    endpoint,
                            COIDMAFence*  fence,
                            uint64_t      dst_handle,
                            off_t   dst_offset,
                            const void*   src_address,
                            off_t   src_offset,
                            off_t   length,
                            COI_COPY_TYPE type = COI_COPY_UNSPECIFIED,
                            void*         task_node = NULL,
                            bool          async=false,
                            uint32_t      frag_count = 0);
};

#endif /* _DMA_H */
