/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#ifndef _ENGINE_H
#define _ENGINE_H
#include <scif.h>
#include <iostream>
#include <list>
#include <stdio.h>
#include <stdlib.h>



#include "../internal/_Debug.h"
#include "../internal/_Message.h"

#include <pthread.h>
#include "../internal/_DynamicDependencyFinderElfStructs.h"

#include "../source/COIProcess_source.h"
#include "../source/COIBuffer_source.h"
#include "../source/COIPipeline_source.h"
#include "../source/COIEvent_source.h"
#include "../source/COIEngine_source.h"
#include "../common/COIMacros_common.h"

// Windows version of pthread types
class _COIProcess;

//=============================================================================
// COIEngine1 Apis -- Start
//=============================================================================

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

// Obtain information (memory available, hw threads, etc)
// about the specified engine.
COIRESULT
COIACCESSAPI
SYMBOL_VERSION(COIEngineGetInfo,1)(
            COIENGINE           in_EngineHandle,
            uint32_t            in_EngineInfoSize,
            COI_ENGINE_INFO*    out_pEngineInfo);

// Get the number of available engines of
// COI_ISA_TYPE in_ISA.
COIRESULT
COIACCESSAPI
SYMBOL_VERSION(COIEngineGetCount,1)(
            COI_ISA_TYPE        in_ISA,
            uint32_t*           out_pNumEngines);

// Get a handle to the nth engine of type in_ISA.
COIRESULT
COIACCESSAPI
SYMBOL_VERSION(COIEngineGetHandle,1)(
            COI_ISA_TYPE        in_ISA,
            uint32_t            in_EngineIndex,
            COIENGINE*          out_pEngineHandle);
#ifdef __cplusplus
}
#endif // __cplusplus

//=============================================================================
// COIEngine1 Apis -- End
//=============================================================================



// This is a copy of the original struct used for backwards compatibility.  It
// would be nice if it wasn't copied from the other current external structure
// but this way the external one doesn't need to expose the internal
// machinations of how the compatibility is handled.
typedef struct COI_ENGINE_INFO_BASE
{
    coi_wchar_t  DriverVersion[COI_MAX_DRIVER_VERSION_STR_LEN];
    COI_ISA_TYPE ISA;
    uint32_t     NumCores;
    coi_eng_misc MiscFlags;
    uint32_t     NumThreads;
    uint32_t     CoreMaxFrequency;
    uint32_t     Load[COI_MAX_HW_THREADS];
    uint64_t     PhysicalMemory;
    uint64_t     PhysicalMemoryFree;
    uint64_t     SwapMemory;
    uint64_t     SwapMemoryFree;
} COI_ENGINE_INFO_BASE;


// Class the represents a Intel® Coprocessor Offload Infrastructure (Intel® COI)  Engine.
class _COIEngine 
{

public:
    _COIEngine();
    ~_COIEngine();

    // Getters for the SCIF node, COI_ISA_TYPE, and ELF64 Machine Type
    // for the Intel® Coprocessor Offload Infrastructure (Intel® COI)  engine (remote)
    uint16_t GetScifNode();
	

    COI_ISA_TYPE GetIsaType();
    Elf64_Ehdr_Machine::Elf64_Ehdr_Machine GetElfMachineType();

    // Getters for the local SCIF node, not the SCIF node where the engine resides
    static uint16_t GetLocalScifNode();
    static uint16_t GetLocalScifNodeUnsafe();

    // Given an COI_ISA_TYPE, returns the number of engines available.
    // Note that just because there is one engine available it doesn't mean that
    // its index is 0. Unfortunately when the API was designed it didn't take
    // offline engines into account and you will have to iterate across
    // all COI_MAX_ISA_MIC_DEVICES to find out which one is the one that is online.
    // TODO - 1) Make this more explicit on the external header
    // TODO - 2) Define a v2 of this API that let's people find out which engines
    //           are online.
    static COIRESULT GetCount(COI_ISA_TYPE isa, uint32_t* out_pNumEngines);

    // Get a handle to the specified COI_ISA_TYPE engine.
    static COIRESULT GetHandle( COI_ISA_TYPE        in_ISA,
                                uint32_t            in_EngineIndex,
                                COIENGINE*          out_pEngineHandle);
    void ForkReset(void);

    // Gets information regarding the engine such as memory available, cpu utilization, etc.
    COIRESULT GetInfo(COI_ENGINE_INFO_BASE* out_pInfo);

    // Some more getters
    inline _SCIFComm& GetComm()
    {
      return m_EngComm;
    }
    inline pthread_mutex_t& GetLock()
    {
      return m_EngLock;
    }

    static pthread_mutex_t& GetClassLock();

    static _COIEngine* Get(COIENGINE engine);
	

    void CleanUpHandler(void);
    COIRESULT DeleteFromProcessList(_COIProcess* process);
    COIRESULT AddToProcessList(_COIProcess* process);
    
protected:
    // Set the SCIF node that the deamon is running on (stored in m_SCIFNode)
    COIRESULT SetScifNode(uint16_t node);
    // Try to connect to the Intel® Coprocessor Offload Infrastructure (Intel® COI)  daemon for this engine
    COIRESULT ConnectToDaemon();

    // SCIF doesn't tell us anything about the architecture of the SCIF nodes.
    // We have some functions that figure it out.
    static uint32_t internal_scif_get_num_x86_64_engines();
    static uint32_t internal_scif_get_num_knf_engines();
    static uint32_t internal_scif_get_num_knc_engines();
    static uint32_t internal_scif_get_num_mic_engines();

private:
    std::list<_COIProcess*> coiProcessList;
    bool m_Connected;    // true when we are connected to the daemon for this engine
    uint16_t m_SCIFNode; // The daemon is running on this SCIF node.
    _SCIFComm m_EngComm; // comm object used by this engine instance
    pthread_mutex_t m_EngLock; // lock for this instance
    pthread_mutex_t m_cleanup_handler_lock; // lock for cleanup handler
	static pthread_mutex_t m_lock; // lock for this class type
    bool m_handler_setup; //boolean for tracking when the CTRL-C handler has been setup so that the handler is set only once
    COI_ISA_TYPE m_ISA;  // COI_ISA_TYPE for this instance
};

#endif /* _ENGINE_H */
