/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#ifndef _ENVHELPER_H
#define _ENVHELPER_H

#include <stdint.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <map>
#include <string>
#include <sstream>
#include "../common/COIResult_common.h"

namespace EnvironmentHelper
{

class EnvHelper
{
public:

static COIRESULT
GetEnv_(const char* const env_var, uint32_t& out_num)
{
    if (NULL == env_var)
    {
        return COI_INVALID_POINTER;
    }

    char* pNodeName = getenv(env_var);
    if (NULL == pNodeName)
    {
        return COI_DOES_NOT_EXIST;
    }

    uint32_t node_num = 0;
    int items_matched = sscanf(pNodeName, "%u", &node_num);

    if (items_matched < 1)
    {
        return COI_ERROR;
    }

    out_num = node_num;

    return COI_SUCCESS;
}

static COIRESULT
GetEnv_(const char* const env_var, uint64_t& out_num)
{
    if (NULL == env_var)
    {
        return COI_INVALID_POINTER;
    }

    char* pNodeName = getenv(env_var);
    if (NULL == pNodeName)
    {
        return COI_DOES_NOT_EXIST;
    }

    uint64_t node_num = 0;
    int items_matched = sscanf(pNodeName, "%lu", &node_num);

    if (items_matched < 1)
    {
        return COI_ERROR;
    }

    out_num = node_num;

    return COI_SUCCESS;
}

};


using std::map;
using std::string;

/* adding an extension to string so that integers (and other later)
 * can be converted to strings automatically.
 * used in the environment vars class */
class mystring : public string {
public:
    using string::operator=;
    string& operator=(const uint64_t i) {
        std::stringstream o;
        o << i;
        (*this) = o.str();
        return *this;
    };
    string& operator=(const uint32_t i) {
        return (*this)=(const uint64_t)i;
    }
    string& operator=(const int i) {
        return (*this)=(const uint64_t)i;
    }
};

/* the creation of the char*[] for env vars seems fairly clumsy
 * so this class uses a map that auto converts to a char*[] */
class environment_map_t : public map<string,mystring> {
public:
    environment_map_t() : tmp_size_(0),tmp_(0) { };
    virtual ~environment_map_t() {
        clear_tmp();
    }

    // Given a string "foo=bar", parse it and add it to the map
    virtual void add(const string& str)
    {
        const char equal = '=';
        size_t pos = str.find(equal);

        // If there wasn't an '=' or it was the located in the final spot
        if( pos == str.npos )
        {
            return;
        }
        if( str.at(str.length()-1) == equal )
        {
            return;
        }

        string left = str.substr(0, pos);
        string right = str.substr(pos+1);
        (*this)[left] = right;
    }

    /* automatically create and convert to an array of char*s */
    operator char**()
    {

        clear_tmp();
        tmp_size_ = (unsigned int) size() + 1;
        tmp_ = new char*[ tmp_size_ ];

        map<string,mystring>::iterator it;
        unsigned int i;
        for(i=0,it=begin();it!=end();++it,++i)
        {
            // 1 for the = and another for the NULL termination char
            size_t str_len = it->first.length() + 2 + it->second.length();
            char* str = new char[str_len+1]; //add one to make klocworks quiet
            snprintf(str,str_len,"%s=%s",it->first.c_str(),it->second.c_str());
            tmp_[i] = str;
        }
        tmp_[size()] = NULL;
        return tmp_;
    };

    operator const char**() {
        return const_cast<const char**>(operator char**());
    }

protected:



    void clear_tmp()
    {
        if (tmp_)
        {
            /* cleans up the allocated char*s */
            for (unsigned int i=0;i<tmp_size_;i++)
            {
                delete[] tmp_[i];
            }
            delete[] tmp_;
        }
        tmp_ = NULL;
        tmp_size_ = 0;
    }

private:
    unsigned int    tmp_size_;
    char**          tmp_;
};

};


#endif /* _ENVHELPER_H */
