/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#ifndef _MESSAGE_H
#define _MESSAGE_H

// Need to include sys/types.h for scif.h to compile.
#include <stdlib.h>
#include <sys/types.h>
#include <scif.h>


#include <memory.h>
#include <memory>
#include <string>
#include <vector>
#include <sys/mman.h>
#include <errno.h>
#include <limits.h>
#include "../common/COIResult_common.h"
#include "../source/COIPipeline_source.h"
#include "../internal/_MemoryRegion.h"
#include "../internal/_PthreadMutexAutoLock.h"


// scif.h no longer does a typedef for its portID
typedef struct scif_portID scif_portID_t;

// Base message class to pass into send and receive below
class Message_t
{
public:
    // default constructor
    Message_t() : m_buffer(NULL), m_size(0) {};

    // copy constructor
    Message_t(const Message_t& rhs)
    {
        deep_copy(rhs); 
    }

    // destructor
    virtual ~Message_t()
    {
        Release();
    }

    // assignment operator
    const Message_t& operator=(const Message_t& rhs)
    {
        Release();
        deep_copy(rhs); 
        return *this;
    }

    // Allocate new space for the message buffer itself.  May free any old
    // data lying around
    char* Allocate(size_t s);
    // Free the buffer, probably only ever used internally
    void Release();

    // Return a pointer to the data that you want sent, minus the "message header"
    char* buffer() const;

    // Return the pointer to the actual data buffer that goes out on the wire
    char* rawdata() const
    {
        return m_buffer;
    }

    // Final number of bytes that the buffer holds
    uint64_t size() const
    {
        return m_size;
    }

protected:
    char*       m_buffer;
    uint64_t    m_size;

private:
    void deep_copy(const Message_t& rhs);
};

// Class whose message is a simple message consisting of a single number, COIRESULT, etc
// If you are using nested op codes that means that message_body actually contains another
// another header and you should use OpcodeMessage_t or one of its subclasses.
template<typename T>
class SimpleMessage_t : public Message_t
{
public:

    T* m_message_body;

    SimpleMessage_t ()
        :
            m_message_body((T*)Allocate(sizeof(T)))
    {
        // Allocate checks and throws exceptions, so we know for sure that
        // m_message_body is not NULL
    }

};

// helper class which starts with an opcode that can be switched off of to
// choose the correct sub sub message type.  see process_source.cpp for an
// example of usage.
class OpcodeMessage_t : public Message_t
{
private:
    // A helper class that lets me use the return type of the GetPayload() in
    // order to do the template differentiation
    class payload_helper
    {
        public:
            payload_helper(const OpcodeMessage_t& m)
                : m_msg(m) {};

            // And then just cast this thing back to the original T* that was
            // on the left hand side of the assignment from the GetPayload()
            // call
            template<typename T>
            operator T*()
            {
                return m_msg;
            }
        private:
            const OpcodeMessage_t&  m_msg;
    };
public:
    virtual uint64_t opcode() const
    {
        return *(uint64_t*)buffer();
    }

    // Allow implicit casting in order to get a sub message payload
    template<typename T>
    operator T*() const
    {
        return (T*)(buffer() + sizeof(uint64_t));
    }

    // Returns a new payload helper that will get cast to the correct type
    // later on
    payload_helper GetPayload() const
    {
        return payload_helper(*this);
    }

    template<typename T>
    void SetPayload(T*& out_payload, int extraspace = 0)
    {
        // Allocate enough space for the new Payload type as well as for the
        // opcode and any extraspace the user needs
        Allocate(sizeof(uint64_t) + sizeof(T) + extraspace);

        // Set the opcode for this buffer from the SUB MESSAGE TYPE's type
        *(uint64_t*)buffer() = T::type;

        // And now update the users's pointer to the new buffer space
        out_payload = this->operator T*();
    }
};

// Macro helper for defining sub messages within an opcode message.
// See _Daemon.h for examples
#define SUB_MESSAGE_TYPE(opcode,elements)   \
    struct opcode ## _T                     \
    {                                       \
        enum { type = opcode };             \
                                            \
        elements                            \
    };                                      \

// *****************************SCIF WRAPPER Classes***************************

// errno 104: Connection reset by peer - the process has died
// errno 107: Not Connected
#define IsErrnoCOI_PROCESS_DIED(err) (err == ECONNRESET || err == ENOTCONN)

class _SCIFComm {

public:
    // Default constructor
    _SCIFComm();
    ~_SCIFComm();

    // This method creates and binds the comm to a given port.
    int Bind(uint16_t portNumber);

    // Connect to given node and portNumber
    int Connect(uint16_t node, uint16_t portNumber, bool reconnect=false);

    // Disconnect the connection.
    int Disconnect();
    void ForkDisconnect();

    //Initialize the comm with already connected endpoint
    int Initialize(scif_epd_t, scif_portID_t);

    // Send an entire message.
    // It will take care of locking against other uses of this comm object.
    COIRESULT SendAtomic(Message_t& message_to_send)
    {
        _PthreadAutoLock_t lock( m_lock );
        return SendUnsafe(message_to_send);
    }

    // COMM functions that end in Unsafe:
    //   Caller must take care of any locking that may be needed.
    // COMM Functions that end in Atomic:
    //   The comm object will use its lock to prevent message mixups.

    // Send an entire message.
    COIRESULT SendUnsafe(Message_t& message_to_send);

    // Receive an entire message.
    COIRESULT ReceiveAtomic(Message_t& message_to_recv)
    {
        _PthreadAutoLock_t lock( m_lock );
        return ReceiveUnsafe(message_to_recv);
    }

    // Receive an entire message.
    // Caller must take care of any locking that may be needed.
    COIRESULT ReceiveUnsafe(Message_t& message_to_recv);

    // Combines the operations
    COIRESULT SendMessageAndReceiveResponseAtomic(Message_t& msg_to_send, Message_t& msg_to_recv)
    {
        _PthreadAutoLock_t lock( m_lock );
        return SendMessageAndReceiveResponseUnsafe(msg_to_send, msg_to_recv);
    }

    COIRESULT SendMessageAndReceiveResponseUnsafe(Message_t& msg_to_send, Message_t& msg_to_recv);

    // Send an array of arrays by first converting it to a single array that
    // is NULL terminated. count indicates that maximum number of entries
    // in the array you wish to send.
    COIRESULT SendStringArrayUnsafe(const char** array, uint32_t count = UINT_MAX);

    // Send files to the other side and gets a response message
    // (COIRESULT) back.
    COIRESULT SendFilesAndRecvResponseUnsafe(std::vector<std::string>& files);

    // Send file buffers to the other side and gets a response message
    // (COIRESULT) back.
    COIRESULT SendFileBuffersAndRecvResponseUnsafe(std::vector<std::string>& names,
                                             std::vector<void*>&       buffers,
                                             std::vector<uint64_t>&    lengths);

    // Receive files and write them to "base_dir", assuming base_dir
    // already exists. Those successfully written are *added* to files_written.
    // Returns COI_SUCCESS only if all of them were written successfully.
    // Also returns the paths of the original files as sent by the sender.
    COIRESULT ReceiveFiles(const std::string& base_dir,
                           /*out*/ std::vector<std::string>& files_written_with_path,
                           /*out*/ std::vector<std::string>& files_written_original);

    // Let the other side know that the connection has closed
    void SendCloseSignal();

    scif_epd_t   GetEndpoint(){ return m_endpoint; }
    scif_portID_t     GetPeerPort(){ return m_peer_port; }

   //Initialize the loop-back or pipe endpoints
    int Initialize_pipe_endpoints(void);

    // The following is used only by Initialize_pipe_endpoints() as a helper function:
    int setupConnection(scif_epd_t inEpd, int portNum);

    // The following type is used only by _temp_thread_static() (following):
    struct _temp_thread_parameter
    {
	scif_epd_t  epd;
	_SCIFComm  *mine;
    };

    // The following is used only by setupConnection() as a helper function:
    static
    void *_temp_thread_static(void *param /* opaque for a _temp_thread_parameter * */)
    {
	_temp_thread_parameter *p = (_temp_thread_parameter *)param;
	return p->mine->_temp_thread(p->epd);
    }

    // The following is used only by _temp_thread_static() (above) as a helper function:
    void *_temp_thread(scif_epd_t);

    // Do not call GetLock() if SendAtomic, ReceiveAtomic, or other "atomic" functions do what you need.
    // Use those instead.
    // Use the GetLock() call with caution.
    
    pthread_mutex_t& GetLock(){ return m_lock; }

protected:
    // scif_send and scif_recv are interruptible, or require spinning (which burns cpu cycles),
    // or require the use of poll() to avoid spinning, etc. These two functions wrap the
    // two scif calls and avoid those issues.
    COIRESULT ScifSendWrapper(void* buffer, size_t size);
    COIRESULT ScifRecvWrapper(void* buffer, size_t size);

    // You need a special page of memory to do a busy loop check
    // for DMA transfer completions.
    COIRESULT AllocateSignalPage();
    // For large enough transfers, a DMA followed by a memcpy
    // will be faster than using scif_send.
    COIRESULT AllocateRegisteredMemory();

    // Local port the scif connection is bound to
    uint16_t    m_port_num;
    // Local scif endpoint we are bound to
    scif_epd_t  m_endpoint;
	// Local scif endpoints used for signalling
	scif_epd_t m_pipe_Recv_endpoint, m_pipe_Send_endpoint;
    // Struct describing the node+port the remote endpoint is bound to
    scif_portID_t    m_peer_port;

    bool        m_initialized;

    // In order to use scif_fence_signal you need to scif_register some
    // page aligned memory.
    // We will need two variables.
    // In order to avoid two checks for what essentially is a single variable,
    // initializations and checks will only be done on one of the two.
    // One could argue that a struct could help out here, but for two variables
    // it almost seems overkill. Almost...
    volatile uint64_t*   m_send_signal_page;
    off_t       m_send_signal_offset;
    // The space we will have to register for scif dma and its corresponding scif offset
    void*  m_registered_space;
    off_t  m_registered_space_offset;

    // Store the remote offset for use in subsequent transfers
    off_t  m_remote_offset;

    
    pthread_mutex_t m_lock;

};


class _SCIFListener {
public:
    // Default Constructor will create, bind, and listen to a random port
    // (portNumber = 0, backlog = 1)
    _SCIFListener();

    //Create, Bind to endpoint and mark it for listen
    _SCIFListener(uint16_t portNumber, int backlog);

    // destructor
    virtual ~_SCIFListener();

    // Wait for the default amount of time for a connection, then time out.
    int WaitForConnect(_SCIFComm& comm);
    // Wait for the specified amount of time for a connection before timing out.
    // Use a negative number to specify "infinite" timeout.
    int WaitForConnect(_SCIFComm&, int timeout_ms);

    // Get the port number we are lisetining on
    int GetPortNumber();

    // Get the scif endpoint we are listening on
    scif_epd_t GetEndpoint(){ return m_endpoint; }

    // Close the listener
    int Close();

    // Bind, and listen to given port number. Returns port number if
    // successful, -1 on error
    int BindAndListen(uint16_t portNumber,int backlog);

private:
    // local scif port number and endpoint we are bound to
    uint16_t m_port_number;
    scif_epd_t m_endpoint;

    bool m_initialized;

    
    pthread_mutex_t m_lock;
};
// ***************************SCIF WRAPPER Classes***************************

#endif /* _MESSAGE_H */
