/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#include <stdlib.h>
#include <memory.h>
#include <sys/types.h>
#include <errno.h>
#include <sched.h>
#include <tr1/memory>
#include <sys/mman.h>
#include <stdint.h>

#include <list>
#include <vector>
#include <algorithm>

#include <common/COIMacros_common.h>
#include <internal/_AutoLock.h>
#include <internal/_Buffer.h>
#include <internal/_DependencyDag.h>
#include <internal/_Log.h>
#include <internal/_Process.h>
#include <internal/_Debug.h>
#include <internal/_MemoryRegion.h>
#include <internal/_PthreadMutexAutoLock.h>
#include <internal/_DMA.h>

#include "buffernodes.h"
#include "pinnedbuffer.h"


COIBufferPinnedBase::COIBufferPinnedBase( uint64_t      size,
                                          uint32_t      num_procs,
                                    const COIPROCESS*   procs,
                                    uint8_t             m_page_type)
    : COIBuffer(COI_BUFFER_PINNED,size,num_procs,procs, (void*)-1, m_page_type)
{
}

COIBufferPinnedBase::~COIBufferPinnedBase()
{
}

uint64_t
COIBufferPinnedBase::RequiredSize()
{
    return 0;
}

COIBufferPinned::COIBufferPinned(         uint64_t      size,
                                          uint32_t      num_procs,
                                    const COIPROCESS*   procs,
                                    uint8_t             m_page_type)
    : COIBufferPinnedBase(size,num_procs,procs, m_page_type),
        m_remapped(false)
{
    // This version only ever has one process, i hope
    ProcessStateInfo* p = FindInfo(procs[0]);

    assert(p!=NULL);

    try {
        if (COI_SUCCESS !=
            p->m_procref->ReserveVirtualSpace(m_actual_size,
                                              &p->m_remote_address, 0))
        {
            throw COI_RESOURCE_EXHAUSTED;
        }

        uint64_t offset =  0;

        // Try to get enough physical blocks for the pinned buffer right now.
        // these are kept for the lifetime of the buffer.  if we can't get
        // enough now then return resource exhausted
        uint64_t need = m_actual_size;
        while (need)
        {
            virtual_region *vr, *prev = NULL;
            physical_region* pr;

            if (COI_SUCCESS != p->m_procref->AllocatePhysicalBlock(this,need,pr))
            {
                throw COI_RESOURCE_EXHAUSTED;
            }
            else
            {
                pr->active = 0;
                pr->refcount = 1;
                vr = new virtual_region(offset,pr->length);
                vr->physical = pr;
                vr->state = COI_BUFFER_VALID;
                need -= vr->length;
                m_physical_blocks.push_back( make_pair(pr,false) );

                if (!prev)
                {
                    LIST_INSERT_HEAD(&p->m_blocks,vr,entries);
                }
                else
                {
                    LIST_INSERT_AFTER(prev,vr,entries);
                }
                prev = vr;
                offset += pr->length;
            }
        }

        // Got all our physical blocks.  so now reserve some local virtual
        // space to map to it
        //
        // Note, no madvise(MADV_DONTFORK) is needed here. Since the buffer
        // is marked MAP_SHARED, the physical mapping will not change.
        m_shadow = mmap(0,m_actual_size,PROT_READ|PROT_WRITE,
                            MAP_NORESERVE|MAP_ANON|MAP_SHARED,-1,0);
        if (MAP_FAILED == m_shadow)
        {
            throw COI_RESOURCE_EXHAUSTED;
        }
        offset = 0;

        // For each physical block, map it into our virtual space locally
        physical_list_t::iterator it;
        for (it=m_physical_blocks.begin(); it!=m_physical_blocks.end(); ++it)
        {
            void* result;
            physical_region* r = it->first;

                result = scif_mmap((void*)((uint64_t)m_shadow + offset),
                                    r->length, SCIF_PROT_READ|SCIF_PROT_WRITE,
                                    SCIF_MAP_FIXED,
                                    p->m_procref->GetEndpoint(),r->offset);
                if ((uint64_t)result != ((uint64_t)m_shadow + offset))
                {
                    throw COI_RESOURCE_EXHAUSTED;
                }
            it->second = true;
            offset += r->length;
        }
    } catch (...) {
        cleanup();
        throw;
    }
}

COIBufferPinned::~COIBufferPinned()
{
    cleanup();
}

void
COIBufferPinned::cleanup()
{
    ProcessStateInfo* p = NULL;
    for(proc_list::iterator it = m_process_info.begin();
        it != m_process_info.end();
        ++it)
    {
        assert(*it);
        if (!(*it)->Shadow())
        {
            p = *it;
        }
    }
    assert(p);
    uint64_t offset = 0;

    // Return all the physical blocks back to the system
    while (!m_physical_blocks.empty())
    {
        elem_t& e = m_physical_blocks.front();

        // if it was mapped locally (and we didn't throw at some point) then
        // unmap it
        if (e.second)
        {
            scif_munmap((void*)((uint64_t)m_shadow + offset), e.first->length);
            offset += e.first->length;
        }

        m_physical_blocks.pop_front();
    }
    if(p!=0)
    {
        p->m_procref->FreeVirtualSpace(m_actual_size,p->m_remote_address);
    }

    if (MAP_FAILED != m_shadow && 0 == offset)
    {
        // Only unmap this region if for some reason it was not scif_unmapped.
        // Otherwise in linux this will unmap memory potentially allocated on a
        // different thread. If offset is zero, then we didn't scif_munmap anything.
        munmap(m_shadow,m_actual_size);
    }
}

COIBufferPinnedFromMemory::COIBufferPinnedFromMemory(
                                uint64_t    size,
                                uint32_t    num_procs,
                          const COIPROCESS* procs,
                                void*       memory,
                                uint8_t     m_page_type)
    : COIBufferPinnedBase(size,num_procs,procs, m_page_type)
{
    m_shadow = memory;
    // For each process register memory and send offset down to be mmaped
    try {
        for (proc_list::iterator it = m_process_info.begin();
            it != m_process_info.end();
            ++it)
        {
            ProcessStateInfo* info = *it;
            assert(*it);
            if (info->Shadow())
            {
                continue;
            }
            _COIProcess* p = info->m_procref;

            // Register shadow memory with scif
            off_t offset, offset_hint = GetNextRegisteredOffsetHint(m_size);
            offset = scif_register(p->GetEndpoint(), m_shadow,
                                   PAGE_CEIL(m_size), offset_hint,
                                   SCIF_PROT_READ | SCIF_PROT_WRITE,
                                   SCIF_MAP_FIXED);
            if (offset < 0)
            {
                throw COI_RESOURCE_EXHAUSTED;
            }
            info->m_shadow_offset = offset;
            info->m_remote_address = 0;

            //Design Note:
            //Pinned buffers never register on dedicated extra DMA endpoints
            //As the are mmaped on the remote side, since mmap returns a single
            //user space virtual address, mmap'ing on multiple endpoints
            //serves no purpose. That is why there is no loop here to register
            //on other endopints.

            // now send offset down so that process can mmap the local memory
            COIRESULT result = p->RemoteRegionMmapSource(offset, size,
                                    &info->m_remote_address);

            if (COI_SUCCESS != result)
            {
                throw result;
            }
        }
    } catch (...) {
        cleanup();
        throw;  // rethrow whatever got us here
    }
}

COIBufferPinnedFromMemory::~COIBufferPinnedFromMemory()
{
    cleanup();
}

void
COIBufferPinnedFromMemory::cleanup()
{
    // clean up the scif registered stuff
    // base class' destruct will get called to clean up the rest
    proc_list::iterator it;
    for (it=m_process_info.begin();it!=m_process_info.end();++it)
    {
        ProcessStateInfo* info = *it;
        assert(*it);
        if (info->Shadow())
        {
            continue;
        }

        if ((uint64_t)-1 != info->m_shadow_offset)
        {   // successfully registered, so unregister
            UNUSED_ATTR int scif_unreg_status = 0;
            scif_unreg_status =
                scif_unregister(info->m_procref->GetEndpoint(),
                            info->m_shadow_offset, PAGE_CEIL(m_size));
            assert(scif_unreg_status == 0 || errno == ENOTCONN);
            info->m_shadow_offset = (uint64_t)-1;

            //Design Note:
            //Pinned buffers only use a single endpoint, so no need
            //to unregister more endpoints, unlike other buffer types.
        }

        // send message down to munmap memory
        info->m_procref->RemoteRegionMunmapSource(info->m_remote_address,
                                                  m_size);
    }
}

COIBufferPinnedMultipleProcs::COIBufferPinnedMultipleProcs(
                    const uint64_t      size,
                    const uint32_t      num_procs,
                    const COIPROCESS*   procs,
                    uint8_t             m_page_type)
try : COIBufferPinnedFromMemory(size,num_procs,procs,
                            posix_memalign_dontfork(m_shadow,PAGE_CEIL(size)), m_page_type)
{
} catch (...) {
    madvise(m_shadow,PAGE_CEIL(size),MADV_DOFORK);
    free(m_shadow);
    throw;
}

COIBufferPinnedMultipleProcs::~COIBufferPinnedMultipleProcs()
{
    cleanup();
}

void
COIBufferPinnedMultipleProcs::cleanup()
{
    // unregister memory
    COIBufferPinnedFromMemory::cleanup();
    madvise(m_shadow,PAGE_CEIL(m_actual_size),MADV_DOFORK);
    free(m_shadow);
}

// Pinned buffers are valid everywhere all the time, so overwrite the moveto
// function so that it doesn't actually do anything.  All that really happens
// is it sets up the buffers events for the range and creates a new refcount
// node.
COIRESULT
COIBufferPinnedBase::MoveTo(
                        COIPROCESS          move_to,
                        COI_ACCESS_FLAGS    flags,
                        uint64_t            offset,
                        uint64_t            length,
                        COIEVENT            in_event,
                        RemapList&          in_remap_list,
                        COIEVENT&           out_complete,
                        COIEVENT&           out_refcount,
                        _COIRunFunction*   run_Function_node)
{
    UNREFERENCED_PARAM(move_to);
    UNREFERENCED_PARAM(flags);
    UNREFERENCED_PARAM(in_remap_list);

    AutoTaskNode<ChokeNode>    choke(new ChokeNode(1));

    ProcessStateInfo*   proc = FindInfo(move_to);

    AddRef(move_to, offset, length, 1);

    choke.AddTask(&in_event);

    out_complete = choke->GetEvent();

    struct buf_region* region = (buf_region*)malloc(sizeof(buf_region));
    if(region == NULL)
    {
        return COI_OUT_OF_MEMORY;
    }
    region->proc = proc;
    region->offset = offset;
    region->length = length;
    region->buf = this;
    run_Function_node->m_regions.insert ( std::pair<_COIBuffer*,buf_region*>(this,region) );

    choke.commit();
    return COI_SUCCESS;
}

COIRESULT
COIBufferPinned::MoveTo(
                        COIPROCESS          move_to,
                        COI_ACCESS_FLAGS    flags,
                        uint64_t            offset,
                        uint64_t            length,
                        COIEVENT            in_event,
                        RemapList&          in_remap_list,
                        COIEVENT&           out_complete,
                        COIEVENT&           out_refcount,
                        _COIRunFunction*   run_Function_node)
{
    COIRESULT result;
    result = COIBufferPinnedBase::MoveTo(move_to, flags, offset, length,
                                         in_event, in_remap_list,
                                         out_complete, out_refcount,
                                         run_Function_node);
    if(COI_SUCCESS != result)
    {
        return result;
    }

    if (!m_remapped)
    {
        m_remapped = true;
        ProcessStateInfo* info = NULL;
        for(proc_list::iterator it = m_process_info.begin();
            it != m_process_info.end();
            ++it)
        {
            assert(*it);
            if (!(*it)->Shadow())
            {
                info = *it;
            }
        }
        assert(info);
        virtual_region* vr;
        if(info!=0)
        {
            LIST_FOREACH(vr,&info->m_blocks,entries)
            {
                AddToRemapList(info,vr,in_remap_list);
            }
        }
    }
    return COI_SUCCESS;
}

COIRESULT
COIBufferPinnedBase::Map( uint64_t          offset,
                          uint64_t          length,
                          COI_MAP_TYPE      type,
                          uint32_t          num_deps,
                    const COIEVENT*         deps,
                          COIEVENT*         out_complete,
                          COIMAPINSTANCE*   out_instance,
                          void**            out_data)
{
    UNREFERENCED_PARAM(type);
    COIEVENT  final_event;

    // preallocate some stuff before locking and working
    MapInstanceImpl* map_instance = new MapInstanceImpl(this,offset,length);

    // pinned buffer uses the buffer_choke_node, since a map of pinned buffer is a 
    // no-op, but it still needs to signal that the "map" has occured.
    AutoTaskNode<buffer_choke_node>   choke(new buffer_choke_node(num_deps, this));

    { // scope the dag Autolock
        TaskScheduler::AutoLock al(TaskScheduler::Get().GetLock());

        { // scope the buffer's lock
            AutoLock al(this);

            if (0 == length)
            {
                length = m_actual_size;
            }

            choke->m_offset      = offset;
            choke->m_length      = length;

            choke.AddTask(deps);
            COIEVENT   input_events[2];
            input_events[0] = choke->GetEvent();

            UpdateEvents(offset, length, true, choke->GetEvent(), input_events[1]);

            m_mappings.push_back(map_instance);

            final_event = choke->GetEvent();

            choke.commit();
        }// end the buffer's lock scope
    }// end of the dag auto lock scope

    // Now set the operations in motion
    TaskScheduler::Get().RunReady();

    if (out_complete)
    {
        *out_complete = final_event;
    }
    else
    {
        TaskScheduler::Get().WaitForEvent(final_event);
    }

    MapInstance::s_valid_mapinstances.Insert(map_instance);

    *out_instance = (COIMAPINSTANCE)map_instance;
    *out_data     = (void*)(((uint64_t)LocalAddress()) + offset);

    return COI_SUCCESS;
}

bool
COIBufferPinnedBase::Mapped()
{
    return false;
}
