/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#include <stdlib.h>
#include <memory.h>
#include <sys/types.h>
#include <errno.h>
#include <sched.h>
#include <tr1/memory>
#include <sys/mman.h>
#include <stdint.h>

#include <list>
#include <vector>
#include <algorithm>

#include <common/COIMacros_common.h>
#include <internal/_AutoLock.h>
#include <internal/_Buffer.h>
#include <internal/_DependencyDag.h>
#include <internal/_Log.h>
#include <internal/_Process.h>
#include <internal/_Debug.h>
#include <internal/_MemoryRegion.h>
#include <internal/_PthreadMutexAutoLock.h>
#include <internal/_DMA.h>

#include "streamingbuffer.h"

_StreamingBuffer::_StreamingBuffer(uint64_t            size,
                           COI_BUFFER_TYPE     type,
                           uint32_t            flags,
                           const void*         init_data,
                           uint32_t            num_procs,
                           const COIPROCESS*   procs,
                           const void*         donated_memory,
                           uint8_t             m_page_type)
    :   COIBufferBase(type,size)
{
    p_initData = NULL;
    m_refCount = 0;

    if(init_data)
    {
        p_initData = malloc(size);
    }

    m_size =  size;
    m_flags = flags;

    if(init_data)
    {
        memcpy(p_initData,init_data,m_size);
    }

    m_numProcs = num_procs;
    p_procs = procs;

    m_type = type;
    donated_memory = donated_memory;

    m_resourceExhausted = true;
    m_usedInPipe = false;

    // Create a NormalFromMem. If Create succeeds, then set
    // resourceExhausted to false;
    p_latestBuffer = _COIBuffer::Create(m_size,COI_BUFFER_NORMAL,m_flags,
            p_initData,m_numProcs,p_procs, NULL);
    _COIBuffer::InsertStream((COIBUFFER)p_latestBuffer);
    m_resourceExhausted = false;
}

COIRESULT
_StreamingBuffer::BufferSetState (
                    COIPROCESS             in_Process,
                    COI_BUFFER_STATE       in_State,
                    COI_BUFFER_MOVE_FLAG   in_DataMove,
                    uint32_t               in_NumDependencies,
            const   COIEVENT*              in_pDependencies,
                    COIEVENT*              out_pCompletion)
{
    UNREFERENCED_PARAM(in_Process);
    UNREFERENCED_PARAM(in_State);
    UNREFERENCED_PARAM(in_DataMove);
    UNREFERENCED_PARAM(in_NumDependencies);
    UNREFERENCED_PARAM(in_pDependencies);
    UNREFERENCED_PARAM(out_pCompletion);
    throw COI_NOT_SUPPORTED;

}


_StreamingBuffer::~_StreamingBuffer()
{
    if(p_initData!=NULL){
	free(p_initData);
    }
}

_COIBuffer*
_StreamingBuffer::CreateSubBuffer(uint64_t, uint64_t)
{
    throw COI_INVALID_HANDLE;
}


void*
_StreamingBuffer::SinkAddress(COIPROCESS p)
{
    return p_latestBuffer->SinkAddress(p);
}

void*
_StreamingBuffer::LocalAddress()
{
    return p_latestBuffer->LocalAddress();
}

void
_StreamingBuffer::MarkUsedInPipe()
{
    m_usedInPipe = true;
}

void
_StreamingBuffer::DeleteVersion(_COIBuffer*)
{
}

_COIBuffer*
_StreamingBuffer::GetCurrentVersion()
{
    return p_latestBuffer;
}

COIRESULT
_StreamingBuffer::Write(    const void*         source,
                            const COIPROCESS    target_process,
                                  uint64_t      offset,
                                  uint64_t      length,
                                  COI_COPY_TYPE type,
                                  uint32_t      num_deps,
                            const COIEVENT*     deps,
                                  COIEVENT*     out_complete,
                                  uint64_t      move_offset,
                                  uint64_t      move_length)
{
    UNREFERENCED_PARAM(source);
    UNREFERENCED_PARAM(offset);
    UNREFERENCED_PARAM(length);
    UNREFERENCED_PARAM(type);
    UNREFERENCED_PARAM(num_deps);
    UNREFERENCED_PARAM(deps);
    UNREFERENCED_PARAM(out_complete);
    return COI_NOT_SUPPORTED;
}

COIRESULT
_StreamingBuffer::MultiDMathCheck(
                            struct arr_desc *   src,
                            struct arr_desc *   dst,
                            uint64_t            off,
                            bool                write)
{
    UNREFERENCED_PARAM(src);
    UNREFERENCED_PARAM(dst);
    UNREFERENCED_PARAM(off);
    UNREFERENCED_PARAM(write);
    return COI_NOT_SUPPORTED;
}

COIRESULT
_StreamingBuffer::WriteMultiD(    
                        const struct arr_desc * source,
                        const COIPROCESS        target_process,
                              uint64_t          offset,
                              struct arr_desc * dest,
                              COI_COPY_TYPE     type,
                              uint32_t          num_deps,
                        const COIEVENT*         deps,
                              COIEVENT*         out_complete,
                              uint64_t          move_offset,
                              uint64_t          move_length)
{
    UNREFERENCED_PARAM(source);
    UNREFERENCED_PARAM(offset);
    UNREFERENCED_PARAM(dest);
    UNREFERENCED_PARAM(type);
    UNREFERENCED_PARAM(num_deps);
    UNREFERENCED_PARAM(deps);
    UNREFERENCED_PARAM(out_complete);
    UNREFERENCED_PARAM(move_offset);
    UNREFERENCED_PARAM(move_length);
    return COI_NOT_SUPPORTED;
}

// Reads from the buffer to a local buffer at dest
COIRESULT
_StreamingBuffer::Read(       void*             dest,
                              uint64_t          offset,
                              uint64_t          length,
                              COI_COPY_TYPE     type,
                              uint32_t          num_deps,
                        const COIEVENT*         deps,
                              COIEVENT*         out_complete)
{
    UNREFERENCED_PARAM(dest);
    UNREFERENCED_PARAM(offset);
    UNREFERENCED_PARAM(length);
    UNREFERENCED_PARAM(type);
    UNREFERENCED_PARAM(num_deps);
    UNREFERENCED_PARAM(deps);
    UNREFERENCED_PARAM(out_complete);

    return COI_NOT_SUPPORTED;
}

COIRESULT
_StreamingBuffer::ReadMultiD(
                    struct arr_desc*    dest,
                    uint64_t            offset,
                    struct arr_desc*    src,
                    COI_COPY_TYPE       type,
                    uint32_t            num_deps,
              const COIEVENT*           deps,
                    COIEVENT*           out_complete)
{
    UNREFERENCED_PARAM(dest);
    UNREFERENCED_PARAM(offset);
    UNREFERENCED_PARAM(src);
    UNREFERENCED_PARAM(type);
    UNREFERENCED_PARAM(num_deps);
    UNREFERENCED_PARAM(deps);
    UNREFERENCED_PARAM(out_complete);

    return COI_NOT_SUPPORTED;
}

COIRESULT
_StreamingBuffer::Copy(     _COIBuffer*     src,
                            uint64_t        dst_offset,
                      const COIPROCESS      target_process,
                            uint64_t        src_offset,
                            uint64_t        length,
                            COI_COPY_TYPE   type,
                            uint32_t        num_deps,
                      const COIEVENT*       deps,
                            COIEVENT*       out_complete,
                            uint64_t      move_offset,
                            uint64_t      move_length)
{
    UNREFERENCED_PARAM(src);
    UNREFERENCED_PARAM(dst_offset);
    UNREFERENCED_PARAM(src_offset);
    UNREFERENCED_PARAM(length);
    UNREFERENCED_PARAM(type);
    UNREFERENCED_PARAM(num_deps);
    UNREFERENCED_PARAM(deps);
    UNREFERENCED_PARAM(out_complete);

    return COI_NOT_SUPPORTED;
}

_ToSinkBuf::_ToSinkBuf(
               uint64_t            size,
               COI_BUFFER_TYPE     type,
               uint32_t            flags,
               const void*         init_data,
               uint32_t            num_procs,
               const   COIPROCESS*   procs,
               const void*         donated_memory,
               uint8_t             m_page_type)
:_StreamingBuffer(size,type,flags,init_data,num_procs,procs,donated_memory, m_page_type)
{
}

_ToSinkBuf::~_ToSinkBuf()
{
}

//Gets called from Pipeline once a RunFunction finishes execution
void
_ToSinkBuf::DeleteVersion(_COIBuffer *p)
{

        TaskScheduler::AutoLock al(TaskScheduler::Get().GetLock());
        AutoLock buf_lock(this);

        //if it is the current version don't delete just return
        if(p == p_latestBuffer)
        {
           m_refCount --; //Decrement Reference count.
           return ;
        }
        m_versions[p] -- ;
        if( m_versions[p] == 0)
        {
           m_versions.erase(p);
           _COIBuffer::Destroy(p);
        }

}

//Destroying all the versions of the buffers
COIRESULT
_ToSinkBuf::Destroy()
{
     TaskScheduler::AutoLock al(TaskScheduler::Get().GetLock());
     AutoLock buf_lock(this);

    COIRESULT result;
    bool cleanedUp = true;
    std::map <_COIBuffer*,uint64_t>::iterator iter;
    //Iterate through all the buffers and call destroy on each version
    for(iter=m_versions.begin(); iter!= m_versions.end();++iter)
    {
        result = _COIBuffer::Destroy((*iter).first);
        if( result != COI_SUCCESS )
        {
            // Need to adjust the bookkeeping for this buffer. The Destroy
            // call would have added the buffer back into the valid buffer
            // map on failure but that map does not track individual versions.
            //
            s_valid_buffers.Remove((*iter).first);            
            if(result == COI_RETRY) //If that version of the buffer is currently mapped
            {
                return result;
            }
            cleanedUp = false;
        }       
    }
    m_versions.clear();

    result = _COIBuffer::Destroy(p_latestBuffer);
    if( result != COI_SUCCESS )
    {
        // Need to adjust the bookkeeping for this buffer. The Destroy
        // call would have added the buffer back into the valid buffer
        // map on failure but that map does not track individual versions.
        //
        s_valid_buffers.Remove(p_latestBuffer);
        if(result == COI_RETRY) //If that version of the buffer is currently mapped
        {
            return result;
        }
        cleanedUp = false;
    }

    if(cleanedUp!=true)
    {
        return COI_ERROR;
    }
    return COI_SUCCESS;
}

COIRESULT
_ToSinkBuf::Move(        COIPROCESS        move_to,
                         COI_ACCESS_FLAGS  flag,
                         uint64_t          offset,
                         uint64_t          length,
                         uint32_t          num_deps,
                   const COIEVENT*         deps,
                         RemapList&        in_remap_list,
                         COIEVENT*         out_complete,
                         COIEVENT*         out_refcount,
                         _COIRunFunction* run_Function_node)
{
    //p_latestBuffer Move will return RETRY if the latest buffer is mapped
    TaskScheduler::AutoLock al(TaskScheduler::Get().GetLock());
    AutoLock buf_lock(this);

    if(m_resourceExhausted || !p_latestBuffer)
    {
        return COI_ERROR;
    }
    else
    {
        m_refCount++;
        return p_latestBuffer->Move(move_to, flag, offset,length,num_deps, deps,
                                in_remap_list,out_complete,out_refcount,run_Function_node);
    }
}

COIRESULT
_ToSinkBuf::Map(        uint64_t          offset,
                        uint64_t          len,
                        COI_MAP_TYPE      type,
                        uint32_t          num_deps,
                        const COIEVENT*   deps,
                        COIEVENT*         out_complete,
                        COIMAPINSTANCE*   out_instance,
                        void**            out_data )
{



    COIRESULT result = COI_ERROR;

    //if the latest version has been used in pipe then only create a new version

    //Why we are not locking the TaskScheduler here?
    //If the map has to wait on the event if releases the taskscheduler lock. Locking
    //the DAG here makes the taskscheduler's lock count as 2 so it doesn't actually release
    //the lock when it waits in pthread_cond_wait and other tasks cannot proceed.
    //As we don't add any task or modify the DAG at this level don't lock the DAG.

    //TODO: Figure out a safer solution so that lock order remains consistent everywhere.

    AutoLock buf_lock(this);

    if(m_usedInPipe)
    {

        _COIBuffer* tmp;
          try {
                tmp = _COIBuffer::Create(m_size,COI_BUFFER_NORMAL,m_flags,
                            p_initData,m_numProcs,p_procs, NULL);
          }
          catch (std::bad_alloc& ) {
                throw COI_OUT_OF_MEMORY;
          } catch (COIRESULT& e) {
                if(e == COI_RESOURCE_EXHAUSTED)
                {
                    m_resourceExhausted = true;
                    tmp = NULL;
                    result = COI_RESOURCE_EXHAUSTED;
                }
         }
        if(m_refCount == 0)
        {
            _COIBuffer::Destroy(p_latestBuffer);
        }
        else
        {
            m_versions[p_latestBuffer]= m_refCount;
        }
        _COIBuffer::InsertStream((COIBUFFER)tmp);
        p_latestBuffer = tmp;
        m_refCount = 0;
        m_usedInPipe = false;
    }

    if(p_latestBuffer)
    {
        m_resourceExhausted = false;
        //Unlock the buffer here. The lock order might end up being buffer->DAG
        //and deletion of the buffer has lock order DAG->buffer. potential deadlock.
        //it is safe to unlock the buffer here, if another thread might change the p_latestbuffer value
        //then both the calls will ultimately be mapping either two different buffers or same buffer twice.
        _COIBuffer* tmp = p_latestBuffer;
        PT_ASSERT(pthread_mutex_unlock(&m_mutex));
        result = tmp->Map(offset,len,type,num_deps,deps,out_complete,
            out_instance,out_data);
        PT_ASSERT(pthread_mutex_lock(&m_mutex));
    }


    return result;
}

bool
_ToSinkBuf::Mapped ()
{
    COIBuffer* b = reinterpret_cast<COIBuffer*>(p_latestBuffer);
    return b->Mapped();
}

_ToSourceBuf::_ToSourceBuf(
               uint64_t            size,
               COI_BUFFER_TYPE     type,
               uint32_t            flags,
               const void*         init_data,
               uint32_t            num_procs,
               const   COIPROCESS*   procs,
               const void*         donated_memory,
               uint8_t             m_page_type)
    :   _StreamingBuffer(size,type,flags,init_data,num_procs,procs,donated_memory, m_page_type)
{
    // add to the list
    m_versions.push_back(p_latestBuffer);
}

_ToSourceBuf::~_ToSourceBuf()
{
}



COIRESULT
_ToSourceBuf::Map(       uint64_t          offset,
                         uint64_t          len,
                         COI_MAP_TYPE      type,
                         uint32_t          num_deps,
                   const COIEVENT*         deps,
                         COIEVENT*         out_complete,
                         COIMAPINSTANCE*   out_instance,
                         void**            out_data)
{

    AutoLock buf_lock(this);

    COIRESULT result = COI_ERROR;

    //Increment reference count of the first buffer
    m_refCount ++;
    _COIBuffer *tmp = m_versions.front();

    //Why we are not locking the TaskScheduler here?
    //If the map has to wait on the event if releases the taskscheduler lock. Locking
    //the DAG here makes the taskscheduler's lock count as 2 so it doesn't actually release
    //the lock when it waits in pthread_cond_wait and other tasks cannot proceed.
    //As we don't add any task or modify the DAG at this level don't lock the DAG.

    //TODO: Figure out a safer solution so that lock order remains consistent everywhere.
    result =  tmp->Map(offset,len,type,num_deps,deps,out_complete,out_instance,out_data);

    if(result == COI_SUCCESS)
    {

        MapInstance* map_instance = new MapStreamingInstanceImpl(this,offset,len,(MapInstanceImpl*)*out_instance,tmp);
        m_mappings.push_back(map_instance);

        *out_instance = (COIMAPINSTANCE)map_instance;
        MapInstance::s_valid_mapinstances.Insert(map_instance);
    }

    return result;
}

COIRESULT
_ToSourceBuf::Unmap(MapStreamingInstanceImpl* map)
{

    COIRESULT result;
    TaskScheduler::AutoLock al(TaskScheduler::Get().GetLock());
    AutoLock buf_lock(this);

    m_refCount--;
    m_mappings.remove(map);
    _COIBuffer *tmp = m_versions.front();
    if(m_refCount==0)
    {
        //Don't erase the last version of the buffer
        if(m_versions.size()== 1)
        {
            return COI_SUCCESS;
        }
        result = _COIBuffer::Destroy(tmp);
        if(result != COI_SUCCESS) //p_latestBuffer points to list's front
        {
            return result;
        }
        m_versions.pop_front();

        //if all the buffers were used up and resource was exhausted p_latestBuffer might be NULL
        //Check if thats the case then try creating a new version here itself as memory got freed up.
        //If it fails , ignore, a version will be created in either next Unmap call or Move Call.
        if(!p_latestBuffer)
        {
            CreateVersion();
        }
    }
    return COI_SUCCESS;
}


COIRESULT
_ToSourceBuf::CreateVersion()
{
    COIRESULT result = COI_SUCCESS;
     try {
            p_latestBuffer = _COIBuffer::Create(m_size,COI_BUFFER_NORMAL,m_flags,
                            p_initData,m_numProcs,p_procs, NULL);
          }
          catch (std::bad_alloc& ) {
                result = COI_OUT_OF_MEMORY;
          } catch (COIRESULT& e) {
                if(e == COI_RESOURCE_EXHAUSTED)
                {
                    m_resourceExhausted = true;
                    p_latestBuffer = NULL;
                    result = COI_RESOURCE_EXHAUSTED;
                }
          }
          //Add to the version list
          if(p_latestBuffer)
          {
            _COIBuffer::InsertStream((COIBUFFER)p_latestBuffer);
            m_versions.push_back(p_latestBuffer);
          }
          return result;
}
COIRESULT
_ToSourceBuf::Move(      COIPROCESS        move_to,
                         COI_ACCESS_FLAGS  flag,
                         uint64_t          offset,
                         uint64_t          length,
                         uint32_t          num_deps,
                   const COIEVENT*         deps,
                         RemapList&        in_remap_list,
                         COIEVENT*         out_complete,
                         COIEVENT*         out_refcount,
                         _COIRunFunction* run_Function_node)
{
    TaskScheduler::AutoLock al(TaskScheduler::Get().GetLock());
    AutoLock buf_lock(this);
    COIRESULT result = COI_ERROR;

    if(Mapped())
    {
        return COI_RETRY;
    }

    //if all the buffers were used up and resource was exhausted p_latestBuffer might be NULL
    //Check if thats the case then try creating a new version
    if(!p_latestBuffer)
    {
       result = CreateVersion();
       if(result != COI_SUCCESS)
       {
           COILOG_FUNC_RETURN_ERROR( result );
       }
    }

    //Do manipulation on the latest buffer
    result = p_latestBuffer->Move(move_to, flag, offset,length,num_deps, deps,
                                in_remap_list,out_complete,out_refcount,run_Function_node);
    if(result != COI_SUCCESS)
    {
           COILOG_FUNC_RETURN_ERROR( result );
    }

    //Create a new version to use it in next Runfunction
    result = CreateVersion();
    if(result != COI_SUCCESS)
    {
           COILOG_FUNC_RETURN_ERROR( result );
    }

    return result;
}

COIRESULT
_ToSourceBuf::Destroy()
{
    TaskScheduler::AutoLock al(TaskScheduler::Get().GetLock());
    AutoLock buf_lock(this);

    COIRESULT result;
    bool cleanedUp = true;

    if (!m_mappings.empty())
    {
        return COI_RETRY;
    }
    std::list <_COIBuffer*>::iterator iter;
    for(iter=m_versions.begin(); iter!= m_versions.end();++iter)
    {
        result = _COIBuffer::Destroy(*iter);
        if( result != COI_SUCCESS )
        {
            // Need to adjust the bookkeeping for this buffer. The Destroy
            // call would have added the buffer back into the valid buffer
            // map on failure but that map does not track individual versions.
            //
            s_valid_buffers.Remove(*iter);
            if(result == COI_RETRY) //If that version of the buffer is currently mapped
            {
                return result;
            }
            cleanedUp = false;
        }
    }
    if(!cleanedUp)
    {
        return COI_ERROR;
    }
    else
    {
        //All versions of the buffer are deleted by Unmap
        return COI_SUCCESS;
    }
}

bool
_ToSourceBuf::Mapped ()
{
    return !m_mappings.empty();
}

MapStreamingInstanceImpl::MapStreamingInstanceImpl(
    _ToSourceBuf*       b,
    uint64_t            offset,
    uint64_t            size,
    MapInstanceImpl*    mapInstance,
    _COIBuffer*         version)
    :   MapInstance(),
        m_buffer(b),
        m_offset(offset),
        m_size(size),
        m_mapping(mapInstance),
        m_version(version)
{
};

MapStreamingInstanceImpl::~MapStreamingInstanceImpl()
{
}

COIRESULT
MapStreamingInstanceImpl::Unmap(    uint32_t          num_deps,
                                    const COIEVENT*   deps,
                                    COIEVENT*         out_complete)
{

    COIRESULT result;
    result = m_mapping->Unmap(num_deps,deps,out_complete);
    MapStreamingInstanceImpl * map = this;
    m_buffer->Unmap(map);
    MapInstance::s_valid_mapinstances.Remove(map);
    delete map;
    map = NULL;
    return result;
}

