/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <scif.h>


#include <string.h>
#include <sstream>

#include <dirent.h>
#include <unistd.h>
#include <stdint.h>
#include <fcntl.h>
#include <signal.h>
#include <exception>
#include <pwd.h>
#include <errno.h>

#include "daemon.h"
#include "util.h"

bool coi_random_user_flag = false;
bool coi_authentication_flag = false;

#define MAX_SCIF_PORT   65536

// loadcalc.cpp
extern void loadcalc_shutdown();

static void signal_handler(int sig, siginfo_t *si, void *)
{
    if (sig == SIGQUIT)
    {
        FATAL("caught %s(%s)\n", strsigsym(sig),
            strsiginfocode(si->si_signo,si->si_code));
    } 
    else
    {
        FATAL("caught %s(%s) on address %p\n", strsigsym(sig),
            strsiginfocode(si->si_signo,si->si_code), si->si_addr);
    }
}

static void print_usage(void)
{
#define FIRST_INDENT " "
#define NORMAL_INDENT "                           "
   fprintf(stderr, "usage: coi_daemon [OPTIONS]\n");
   fprintf(stderr, "OPTIONS:\n");
   fprintf(stderr, "  %-24s"
           FIRST_INDENT
           "This option enables output from the daemon and\n"
           NORMAL_INDENT
           "child (sink) processes without a proxy to be sent\n"
           NORMAL_INDENT
           " to <stream>. Without this option output gets sent\n"
           NORMAL_INDENT " to /dev/null\n"
           NORMAL_INDENT
           "--enable-output is a long alias for --\n"
           , "--[=<stream>]");

   fprintf(stderr, "\n");
   fprintf(stderr, "  %-24s"
           FIRST_INDENT
           "The maximum number of connections we allow from host\n"
           NORMAL_INDENT
           "processes. If this is exceeded, new connections\n"
           NORMAL_INDENT
           "are temporarily blocked. Defaults to %d.\n"
           , "--max-connections=<int>", DEFAULT_MAX_CONNECTIONS);

   fprintf(stderr, "\n");
   fprintf(stderr, "  %-24s"
           FIRST_INDENT
           "A valid system username. If this is used the coi_daemon\n"
           NORMAL_INDENT
           "will set priviledges to this user after startup.\n\n"
           NORMAL_INDENT
           "Specifying '%s' will create a randomized user for each process.\n\n"
           NORMAL_INDENT
           "Specifying '%s' will enable authentication.\n"
           NORMAL_INDENT
           "The coi_daemon will attempt to validate users who use the coi_daemon.\n"
           NORMAL_INDENT
           "Unauthorized users will be unable to use the coi_daemon,\n"
           NORMAL_INDENT
           "and processes created by the coi_daemon will have the permissions\n"
           NORMAL_INDENT
           "of the user that requested its creation.\n"
           NORMAL_INDENT
           "Currently, this mode is not supported in Windows.\n\n"
           NORMAL_INDENT
           "Defaults to '%s'.\n"
           , "--coiuser=<username>", "_Dynamic", "_Authorized", DEFAULT_USER);

   fprintf(stderr, "\n");
   fprintf(stderr, "  %-24s"
       FIRST_INDENT
       "A valid port number. If this is used the coi_daemon will attempt\n"
       NORMAL_INDENT
       "to bind to this port. Additionally COI_DAEMON_PORT needs to be\n"
       NORMAL_INDENT
       "exported on the host with this same value so the host application\n"
       NORMAL_INDENT
       "will be able connect with the coi_daemon. Defaults to '%d'.\n"
       , "--coiport=<port>", (int)COI_DAEMON_PORT);

   fprintf(stderr, "\n");
   fprintf(stderr, "  %-24s"
           FIRST_INDENT
#ifdef ENABLE_LOGGING
           "Where to write the log file to. See the <stream>\n"
           NORMAL_INDENT
           "definition below.\n"
           NORMAL_INDENT
           "indicates stdout, - means stderr. By default no\n"
           NORMAL_INDENT
           "log is generated.\n"
#else
           "Logging disabled, this argument will be ignored.\n"
#endif
           , "--log[=<stream>]");
    fprintf(stderr, "\n  %-24s"
        FIRST_INDENT "Can be an existing stream, a file or socket\n"
        NORMAL_INDENT "No argument means to use stdout. An argument of -\n"
        NORMAL_INDENT "means stderr. A path indicates a file and a\n"
        NORMAL_INDENT "host:port pair opens a socket.\n"
        NORMAL_INDENT "EXAMPLES:\n"
        NORMAL_INDENT " --               (stdout)\n"
        NORMAL_INDENT " --=-             (stderr)\n"
        NORMAL_INDENT " --=/tmp/bar.log  (a file)\n"
        NORMAL_INDENT " --=host:27001    (a socket).\n"
        NORMAL_INDENT "        E.g: first run (nc -l 27001) on host\n"
        , "<stream>"
        );
    fprintf(stderr, "\n");
}

// atoi and strtol admit some invalid input e.g. atoi("123asdkfj")
// Neither reports errors, they simply return 0.
static bool parse_int(const char *arg, int *i)
{
    int sign = 1;
    int x = 0;
    int n = 0;

    if(!arg) 
    {
        WARN ("%s: arg was NULL\n", __FUNCTION__);
        return false;
    }

    if(!i)
    {
        WARN ("%s: i was NULL\n", __FUNCTION__);
        return false;
    }

    if (*arg == '-')
    {
        arg++;
        sign = -1;
    }

    if ( sscanf(arg,"%d%n",&x,&n) == 0 || 
            ((unsigned int)n < strlen(arg) &&
            arg[n] != 0))
    {
        return false;
    }

    *i = sign * x;

    return true;
}

static bool parse_sock_stream(const char *str, struct sockaddr_in *a)
{
    int port = -1;
    const char *portstr = NULL;

    if(!str) 
    {
        WARN("%s: str was NULL\n", __FUNCTION__);
        return false;
    }

    if(!a) 
    {
        WARN("%s: a was NULL\n", __FUNCTION__);
        return false;
    }

    portstr = strchr(str,':');
    if (portstr == NULL || !parse_int(portstr+1, &port) ||
        port < 0 || port > MAX_SCIF_PORT)
    {
        return false;
    }

    char host[256] = { 0 };
    unsigned len = portstr - str;
    if ((unsigned)len >= sizeof(host) - 1) {
        return false;
    }
    strncpy(host, str, len);
    host[len + 1] = '\0';
    return (0 == resolve_addr(host, (unsigned short)port, a));
}


// Parses a stream argument and opens or redirects a file descriptor for
// that stream. Examples:
// --foo                  (no arg) means use STDOUT_FILENO
// --foo=-                - mean suse STDERR_FILENO
// --foo=file             file means use file
// --foo=somehost:27001   open a socket to somehost on port 27001
//
// val points to whatever follows the = (or NULL or the empty string
// "--foo=file"                          for no argument)
//        ^  val
//
// The dup_stdio flag says that if the stream used is a stdio stream, dup it.
int openstream(const char *val, bool dup_stdio)
{
    if (val == NULL || *val == 0) {
        if(dup_stdio) {
            return dup(STDOUT_FILENO);
        } else {
            return STDOUT_FILENO;
        }
    } else if (strcmp(val,"-") == 0) {
        if(dup_stdio) {
            return dup(STDERR_FILENO);
        } else {
            return STDERR_FILENO;
        }
    } else {
        struct sockaddr_in raddr;
        if (parse_sock_stream(val,&raddr)) {
            return connect_socket(&raddr, 2000);
        } else {
            return open(val, O_WRONLY|O_CREAT|O_TRUNC, 0666);
        }
    }
}


int main(int argc, char **argv)
{
    char coi_user[MAX_USERNAME_LENGTH];
    bool coi_user_set = false;
    int bCustomUser = 0;
    int max_connections = DEFAULT_MAX_CONNECTIONS;
    uint16_t coi_port = 0;
    int output_fd = -1;

    for (int ai = 1; ai < argc; ai++)
    {
        const char *arg = "";
        if (strprefix(argv[ai],"--help",NULL))
        {
            print_usage();
            return -1;
        }
        else if (strprefix(argv[ai],"--max-connections=",&arg))
        {
            if (!parse_int(arg,&max_connections)) {
                fprintf(stderr, "malformed connection argument: \'%s\'\n", arg);
                exit(-1);
            }
        }
        else if (strprefix(argv[ai],"--coiuser=",&arg))
        {
            if(coi_user_set)
            {
                fprintf(stderr, "Cannot set permissions of a process to multiple users!\n");
                return -1;
            }
            coi_user_set = true;

            size_t usersize = strlen(arg);
            if((usersize > 0) && (usersize < MAX_USERNAME_LENGTH)) {
                if(strncmp(arg, "_Dynamic", sizeof(coi_user)-1)==0){
					coi_random_user_flag = true;
                }
                else if(strncmp(arg, "_Authorized", sizeof(coi_user)-1)==0){
                    coi_authentication_flag = true;
                }
                else if(strncpy(coi_user, arg, sizeof(coi_user)-1))
                {
                    bCustomUser = 1;
                    coi_user[sizeof(coi_user)-1] = 0;
                }
                else
                {
                    fprintf(stderr, "strncpy failed to get custom user\n");
                }
            } else {
                fprintf(stderr, "'%s': is an invalid username, terminating\n",
                        arg);
                print_usage();
                return -1;
            }
        }
        else if (strprefix(argv[ai],"--coiport=",&arg))
        {
            stringstream ss(arg);
            ss >> coi_port;
            //ss returns null if the first term of coi_port isn't a number.
            //ss ignores everything that isn't a number past the first term
            //ss returns null if the number is > 65535
            //ss reroutes negatives to be added to 65535
            //don't allow the use of port 0.
            if (!ss || coi_port < 1)
            {
                fprintf(stderr,
                        "'%s': is an invalid port number, terminating\n",
                        arg);
                return -1;
            }
        }
        else if (strcmp(argv[ai],"--") == 0 ||
            strcmp(argv[ai],"--enable-output") == 0 ||
            strprefix(argv[ai],"--=", &arg) ||
            strprefix(argv[ai],"--enable-output=", &arg))
        {
            if (output_fd != -1) {
                fprintf(stderr, "-- option specified multiple times\n");
                return -1;
            } else if ((output_fd = openstream(arg,false)) == -1) {
                fprintf(stderr, "%s: failed to open stream: %s\n",
                    argv[ai], strerror(errno));
                return -1;
            }
        }
        else if (strcmp(argv[ai],"--log") == 0 ||
            strprefix(argv[ai],"--log=",&arg))
        {
            int fd = -1;
#ifdef ENABLE_LOGGING
            if (g_log_file != NULL) {
                fprintf(stderr, "--log option specified multiple times\n");
                return -1;
            } else if ((fd = openstream(arg,true)) == -1 ||
                (g_log_file = fdopen(fd,"w")) == NULL)
            {
                if(arg!=0)
                {
                    fprintf(stderr, "--log failed to open %s: %s\n", arg,
                    strerror(errno));
                }
                return -1;
            }
            setCLOEXEC(fd);
#else
            const char *errmsg = "ignoring --log argument. Compile"
                                " with ENABLE_LOGGING to enable output\n";
            fprintf(stderr, errmsg);

            if(!g_log_file) {
                if ((fd = openstream(arg, false)) == -1 || 
                     (g_log_file = fdopen(fd,"w")) == NULL)
                {
                    if(arg!=0)
                    {
                        fprintf(stderr, "--log failed to open %s: %s\n", arg,
                        strerror(errno));
                    }
                    if((fd > -1) && close(fd)) //potentially this was opened
                    {
                        fprintf(stderr, "Failed to close log file\n");
                        return -1;
                    }
                } else {
                    // in case they go looking for their log file
                    fprintf(g_log_file, errmsg); 
                    fflush(g_log_file);
                    fclose(g_log_file);
                    g_log_file=NULL;
                    if(close(fd))
                    {
                        fprintf(stderr, "Failed to close log file\n");
                        return -1;
                    }
                }
            } else {
                fprintf(stderr, "--log option specified multiple times\n");
                return -1;
            }   
#endif
        }
        else
        {
            fprintf(stderr, "unrecognized argument \'%s\'\n",argv[ai]);
            print_usage();
            return -1;
        }
    }

    DIR* dir = opendir("/tmp/COI2MB");
    if (dir)
    {
        closedir(dir);
    }
    else if (ENOENT == errno)
    {
        mkdir("/tmp/COI2MB", S_ISVTX | S_IRWXU | S_IRWXG | S_IRWXO);
        system("mount -t hugetlbfs none /tmp/COI2MB/");
        system("chmod 777 /tmp/COI2MB");
    }

    // Open the listener first. This gives the user a much better error message
    // for a common problem: the daemon is already running.
    // Also this must occur BEFORE the setcoiuser call, in case we are binding
    // to a priviledged port, which is always the case at the moment.
    _SCIFListener *listener = NULL;
    try {
	if(coi_port != 0) {
	    listener = new _SCIFListener(coi_port,max_connections);
	} else {
            listener = new _SCIFListener(COI_DAEMON_PORT,max_connections);
	}
    } catch (COIRESULT r) {
        // We don't need to dump diagnostic information for this error.
        fprintf(stderr, "Exception opening listener:\n  %s (%s): Is the daemon already"
            " running?\n", COIRESULTStr(r), strerror(errno));
        exit(1);
    }

    // Drop permissions here to specified user, we don't want to open
    // any file descriptors while retaining super user priviledges
    if ( bCustomUser ) {
        if(setcoiuser(coi_user)) {
            fprintf(stderr, "Can't drop priviledges to requested user '%s'\n",
                coi_user);
            return -1;
        }
    }
    else
    {
        if(setcoiuser(DEFAULT_USER)) {
            fprintf(stderr, "Can't set user to default user: %s\n", DEFAULT_USER);
            return -1;
        }
    }

    // Figure out the scif node we are running on
    if (scif_get_nodeIDs(NULL, 0, &g_scif_node) < 0) {
        FATAL("Unable to get an accurate count of total number of "
                        "scif nodes in the system\n");
    } else if (g_scif_node == (uint16_t)-1) {
        FATAL("Unable to determine local scif node that is being "
                        "used to run this process\n");
    }

    // If the proxy is not enabled, children inherit our stdout and stderr.
    // Since the daemon runs without a tty, when the children sinks try and
    // read or write, they will die from a broken pipe. Hence, we set our
    // stdout and stderr to /dev/null for them to inherit unless the user
    // explicitly demands output via the --enable-output option.
    int dev_null = -1;
    if (output_fd == -1) {
        if ((output_fd = dev_null = open("/dev/null",O_RDWR)) == -1) 
        {
            FATAL("failed to open /dev/null: %s\n", strerror(errno));
        }
        g_headless = true;
    }

    if (dup2(output_fd, STDIN_FILENO) == -1) {
        FATAL("dup2(.., STDIN_FILENO): %s", strerror(errno));
    }
    if (dup2(output_fd, STDOUT_FILENO) == -1) {
        FATAL("dup2(.., STDOUT_FILENO): %s", strerror(errno));
    }
    if (dup2(output_fd, STDERR_FILENO) == -1) {
        FATAL("dup2(.., STDERR_FILENO): %s", strerror(errno));
    }
    if (dev_null != -1) {
        if(close(dev_null))
        {
            FATAL("close(dev_null): %s", strerror(errno));
        }
    }
    if (getenv("COI_LIBRARY_PATH"))
    {
        COI_LIB = getenv("COI_LIBRARY_PATH");
        
    }
    else
    {
        // Will fail if -DCOI_LIB_VERSION wasn't added at some point to the
        // compile defines. It should be in the top most CMakeLists.txt
        mystring coi_lib_version;
        coi_lib_version = COI_LIB_VERSION;
        COI_LIB = "libcoi_device.so";
        COI_LIB += ".";
        COI_LIB += coi_lib_version;
    }

    INFO("--- Intel® Coprocessor Offload Infrastructure (Intel® COI)  DAEMON STARTED ---\n");
    INFO("  pid: %d\n", getpid());
    INFO("  built: %s %s\n", __DATE__, __TIME__);
    INFO("  lib-version: %d\n", COI_LIB_VERSION);
    INFO("  Intel® Coprocessor Offload Infrastructure (Intel® COI)  API VERSION: %s\n", COI_CONNECTION_API_VERSION_STR);

    //  Pin the daemon to the BSP core.
    cpu_set_t *cpumask = get_daemon_cpu_set();
    if (sched_setaffinity(0, sizeof(*cpumask), cpumask) != 0) {
        WARN("sched_setaffinity: %s\n", strerror(errno));
    }

    // Install a fatal handler for SIGSEGV and SIGBUS'es
    struct sigaction sa;
    sa.sa_sigaction = signal_handler;
    sigemptyset(&sa.sa_mask);
    sigaddset(&sa.sa_mask,SIGINT);
    sa.sa_flags = SA_SIGINFO;
    if (sigaction(SIGSEGV,&sa,NULL) ||
        sigaction(SIGBUS,&sa,NULL) ||
        sigaction(SIGQUIT,&sa,NULL))
    {
        FATAL("Failed to install signal handler %s\n", strerror(errno));
    }

    Daemon d(max_connections);
    g_daemon = &d;

    // Execute the main loop to listen for, and process, connections.
    try {
        d.MainLoop(*listener, coi_port);
    } catch (std::exception& e) {
        FATAL("Exception in MainLoop: (%s):\n", e.what());
    } catch (const char *err) {
        FATAL("Exception in MainLoop: (%s):\n", err);
    } catch (COIRESULT &r) {
        FATAL("Exception in MainLoop: %s\n", COIRESULTStr(r));
    } catch (...) {
        FATAL("Unknown exception in MainLoop\n");
    }

    loadcalc_shutdown();

    delete listener;

    if (g_log_file != NULL) {
        fflush(g_log_file);
    }

    g_daemon = NULL;

    return 0;
}

