/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#include <stdlib.h>
#include <string.h>
#include <sstream>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>

#include <unistd.h>
#include <pwd.h>

#include <scif.h>



#include <internal/_Engine.h>
#include <internal/_Daemon.h>
#include <internal/_PthreadMutexAutoLock.h>
#include <internal/_Debug.h>
#include <internal/_Log.h>
#include <internal/_Daemon.h>
#include <source/COIEngine_source.h>
#include <internal/_Process.h>
#if 0
#define DPRINTF(...) printf(__VA_ARGS__)
#else
#define DPRINTF(...)
#endif

pthread_mutex_t _COIEngine::m_lock = PTHREAD_MUTEX_INITIALIZER;

//This indicates "host" as a physical destination
#define SCIF_HOST_DEST 0

// These are where all the instances of _COIEngine's will be.
// To get a COIENGINE handle, you only need a typecast.
// TODO: One could make the argument that you could have some sort of map
// and do some type of initialization somewhere for it. May be overcomplex. Discuss.
static _COIEngine g_engines_x86_64[COI_MAX_ISA_x86_64_DEVICES];
static _COIEngine g_engines_KNF[COI_MAX_ISA_KNF_DEVICES];
static _COIEngine g_engines_KNC[COI_MAX_ISA_KNC_DEVICES];

#ifdef COIHOST
static bool g_AtForkRegistered = false;
#endif

void engine_fork_handler(void)
{
    COIRESULT coi_result;
    COIENGINE pEngine;
    for(int i=0; i < COI_MAX_ISA_KNC_DEVICES; i++)
    {
        coi_result = _COIEngine::GetHandle(COI_ISA_KNC, i, &pEngine);
        if(coi_result == COI_SUCCESS)
        {
            g_engines_KNC[i].ForkReset();
        }
    }
}
void EngineCleanUpHandler(int signum)
{

    COIRESULT coi_result;
    COIENGINE pEngine;
    for(int i=0; i < COI_MAX_ISA_KNC_DEVICES; i++)
    {
        coi_result = _COIEngine::GetHandle(COI_ISA_KNC, i, &pEngine);
        if(coi_result == COI_SUCCESS)
        {
            g_engines_KNC[i].CleanUpHandler();
        }
    }
	exit(signum);
}

// This function is safe betwen windows and linux
// When we find an equivalent pthread_atfork handler for windows
// we can use this function to protect the daemon from forks()
void _COIEngine::ForkReset(void)
{
    if(m_Connected)
    {
         _PthreadAutoLock_t _l(m_EngLock);
         m_EngComm.ForkDisconnect();
         m_Connected = false;
    }
}

	

pthread_mutex_t& _COIEngine::GetClassLock()
{
    static pthread_mutex_t g_EngineClassInitLock = PTHREAD_MUTEX_INITIALIZER;
    return g_EngineClassInitLock;
}

_COIEngine::_COIEngine() :
                m_Connected(false),
                m_SCIFNode((uint16_t)-1)
{
    // No need to set this mutex as re-entrant.
    m_handler_setup=false;
    pthread_mutexattr_t   mta;
    pthread_mutexattr_init(&mta);
    pthread_mutexattr_settype(&mta, PTHREAD_MUTEX_DEFAULT);

    pthread_mutex_init(&m_EngLock,&mta);
    pthread_mutex_init(&m_cleanup_handler_lock,&mta);
    pthread_mutexattr_destroy(&mta);
#ifdef COIHOST
    if(!g_AtForkRegistered)
    {
        g_AtForkRegistered = true;
        //These registered handlers are stackable, so we want
        //to make sure they get called only one time.
        pthread_atfork(NULL, NULL, &engine_fork_handler);
    }
#endif
    
	
}

_COIEngine::~_COIEngine()
{
    if (m_Connected)
    {
        m_EngComm.Disconnect();
    }
}

//Ask the mpss daemon to create ~/.mpsscookie on host and card, which may be used for authentication.
//Returns 0 on success, and something else on error.
//Only works in Linux.
unsigned int create_mpsscookie()
{
    //A port from the host to the MPSS host daemon.
    struct scif_portID mpssdID = {SCIF_HOST_DEST, MPSSD_CRED};
    uid_t uid = getuid();
    scif_epd_t mon_ep;
    struct scif_pollepd parr;
    unsigned int proto;

    if ((mon_ep = scif_open()) < 0)
    {
        COILOG_ERROR("Failed to open SCIF: %s\n", strerror(errno));
        return -1;
    }

    if (scif_connect(mon_ep, &mpssdID) < 0)
    {
        COILOG_ERROR("Faied to connect to mpssd monitor thread on host: %s\n", strerror(errno)); 
        return -1;
    }

    if (scif_send(mon_ep, &uid, sizeof(uid), SCIF_SEND_BLOCK) < 0)
    {
        COILOG_ERROR("Failed to send start packet to host: %s\n", strerror(errno));
        scif_close(mon_ep);
        return -1;
    }

    //Wait for up to five seconds for the mpssd to return information.
    parr.epd = mon_ep;
    parr.events = SCIF_POLLIN;
    parr.revents = 0;
    if(scif_poll(&parr, 1, 5000.0) < 0)
    {
        COILOG_ERROR("Polling for mpssd daemon failed: %s\n", strerror(errno));
        return -1;
    }
    if(scif_recv(mon_ep, &proto, sizeof(proto), 0) < 0)
    {
        COILOG_ERROR("Failed to receive from the mpssd daemon: %s\n", strerror(errno));
        return -1;
    }

    return proto;
}

//Finds and sends an authentication key from ~/.mpsscookie
//This key was provided by the MPSS_d, and might be used for authentication.
//Authentication currently only works for Linux, and only if a flag is set.
//Silently fails, since we don't know if authorization will be necessary
//until we get to the card. The host_key_size will be zero if any problems are
//encountered.
void get_auth_key(char *username, char **host_key, uint64_t* host_key_size, std::string cookie_path)
{
    int host_fd;
    uint32_t read_count = 0;
    int32_t read_ret;
    struct stat key_stat;
    //Max file size to prevent OOM attack
    const int max_limit = 1024 * 1024;

    if(username == NULL || host_key_size == NULL)
    {
        return;
    }
    *host_key_size = 0;

    if(create_mpsscookie() != 0)
    {
        return;
    }

    if ((host_fd = open(cookie_path.c_str(), O_RDONLY)) == -1)
    {
        return;
    }

    //Get key size
    if (fstat(host_fd, &key_stat))
    {
        close(host_fd);
        return;
    }

    if (key_stat.st_size > max_limit)
    {
        close(host_fd);
        return;
    }

    //Malloc host_key to file size plus one to ensure eof is hit when we read
    if (*host_key)
        free(*host_key);
    *host_key = (char *)malloc(sizeof(char) * (key_stat.st_size + 1));
    if (*host_key == NULL)
    {
        COILOG_ERROR("Could not malloc memory for host key\n");
        close(host_fd);
        return;
    }

    // Read until EOF is encountered and we have read the expected number of bytes
    while ((read_ret = read(host_fd, *host_key + read_count, key_stat.st_size - read_count + 1))
           && read_count < key_stat.st_size)
    {
        read_count += read_ret;
        // Fail if there is a reading error or if we read an incorrect size of bytes
        if (read_ret <= 0 || read_count > key_stat.st_size)
        {
            close(host_fd);
            return;
        }
    }

    *host_key_size = read_count;
    close(host_fd);
}

COIRESULT
_COIEngine::ConnectToDaemon()
{
    COILOG_FUNC_ENTER;

    if( m_Connected )
    {
        COILOG_FUNC_RETURN_RESULT(COI_ALREADY_INITIALIZED);
    }

    COIRESULT result = COI_ERROR;
    int status;
    COIDaemonMessage_t message;
    COIDaemonMessage_t::DAEMON_CONNECTION_REQUEST_T* pRequest;
    COIDaemonMessage_t::DAEMON_CONNECTION_RESULT_T* pResponse;
    char *username = NULL; //We can always allocate this for now even for windows
    //For authentication, which is currently only set up for linux
    char *auth_key = NULL;
    uid_t uid;
    passwd* p;
    uint64_t auth_key_size = 0;
    const char *coi_port_env;

    status = m_EngComm.Bind(0);
    if (-1 == status)
    {
        COILOG_ERROR("Engine_bind");
        goto end;
    }

    coi_port_env = getenv("COI_DAEMON_PORT");

    if (coi_port_env == 0 || coi_port_env[0] == '\0')
    {
        status = m_EngComm.Connect(m_SCIFNode, COI_DAEMON_PORT);
        if (-1 == status)
        {
            COILOG_ERROR("Engine_connect");
            if( errno == ENODEV )
            {
                result = COI_NOT_INITIALIZED;
            }
            goto end;
        }
    }
    else
    {
        uint16_t coi_port = 0;
        stringstream ss(coi_port_env);
        ss >> coi_port;
        //ss returns null if the first term of coi_port isn't a number.
        //ss ignores everything that isn't a number past the first term.
        //ss returns null if the number is > 65535
        //ss reroutes negatives to be added to 65535
        //don't allow the use of port 0.
        if (!ss || coi_port < 1)
        {
            COILOG_ERROR("$COI_DAEMON_PORT contains invalid port");
            result = COI_BAD_PORT;
            goto end;
        }
        else
        {
            status = m_EngComm.Connect(m_SCIFNode, coi_port);
            if (-1 == status)
            {
                COILOG_ERROR("Engine_connect");
                if( errno == ENODEV )
                {
                    result = COI_NOT_INITIALIZED;
                }
                goto end;
            }
        }
    }

    COILOG_INFO("Connection to Daemon established. Still need to complete engine info handshake.");

    //Currently, only Linux is set up to support authentication.
    uid = getuid();
    p = getpwuid(uid);
    if(p != NULL)
    {
        std::string host_key_path;
        username = p->pw_name;
        if(p->pw_uid == 0)
        {
            host_key_path = "/" + std::string(username) + "/.mpsscookie";
        }
        else
        {
            host_key_path = "/home/" + std::string(username) + "/.mpsscookie";
        }
        get_auth_key(username, &auth_key, &auth_key_size, host_key_path);
    }

    //Allocate the request message.
    message.SetPayload(pRequest, (int) sizeof(char)*((int) auth_key_size));
    strncpy(pRequest->sink_version, COI_CONNECTION_API_VERSION_STR, sizeof(pRequest->sink_version));
    pRequest->sink_version[sizeof(pRequest->sink_version)-1] = 0;
    if(!username)
    {
        //If we didn't set username up, then just clear the message field
        memset(&pRequest->username, '\0', MAX_USERNAME_LENGTH);
    }
    else
    {
        //Else copy the real username into the message
        memcpy(&pRequest->username, username, MAX_USERNAME_LENGTH);
    }
    pRequest->key_size = auth_key_size;
    memcpy(&pRequest->key[0], auth_key, pRequest->key_size);
    if(auth_key) free(auth_key);
    pRequest->host_has_windows = false;

    //Send the request message and receive the result message.
    result = m_EngComm.SendMessageAndReceiveResponseAtomic(message, message);
    if(result != COI_SUCCESS)
    {
        goto end;
    }
    pResponse = message.GetPayload();

    //Parse the response message.
    if(pResponse->result == COI_AUTHENTICATION_FAILURE)
    {
        result = COI_AUTHENTICATION_FAILURE;
        COILOG_ERROR("Connected to the remote engine and "
                     "it was unable to authenticate the user.\n"
                     "Please contact your administrator "
                     "for more information.\n");
        goto end;
    }
    else if(pResponse->result != COI_SUCCESS)
    {
        result = COI_VERSION_MISMATCH;
        //Make sure the string is terminated before printing it out.
        pResponse->sink_version[COI_CONNECTION_API_VERSION_STR_MAX_SIZE-1] = '\0';
        COILOG_ERROR("Connected to the remote engine and "
                     "it indicated that it supports COI version %s, "
                     "but the local side was compiled with version %s.\n",
                     pResponse->sink_version,
                     COI_CONNECTION_API_VERSION_STR);
        goto end;
    }

    COILOG_INFO("EngineInfo setup complete");
    result = COI_SUCCESS;
    m_Connected = true;

end:
    if( result != COI_SUCCESS)
    {
        m_EngComm.Disconnect();
        m_Connected = false;
    }
    COILOG_FUNC_RETURN_RESULT(result);
}

uint16_t _COIEngine::GetLocalScifNode()
{
    _PthreadAutoLock_t lock( _COIEngine::GetClassLock());
    return GetLocalScifNodeUnsafe();
}

uint16_t _COIEngine::GetLocalScifNodeUnsafe()
{
    static uint16_t self = (uint16_t)-1;

    if( self == (uint16_t)-1 )
    {
        UNUSED_ATTR int count;
        // Work-around scif_get_nodeIDs failure when passing NULL
        uint16_t ignore[2];
        count = scif_get_nodeIDs(ignore, 2, &self);
        assert( count > 0 );
        assert( self != (uint16_t)-1);
    }

    return self;
}
uint16_t
_COIEngine::GetScifNode()
{
    return m_SCIFNode;
}

// Let's connect as soon as we know what our
// SCIF node is. This allows us to avoid checking
// whether or not we are connected from here on out.
COIRESULT
_COIEngine::SetScifNode(uint16_t node)
{
    COIRESULT result = COI_SUCCESS;
    _PthreadAutoLock_t _l(m_EngLock);

    // TODO - Re-evaluate if you should to connect to yourself once loop-back is enabled.
    if( node == GetLocalScifNodeUnsafe() )
    {
        m_SCIFNode = node;
        return COI_SUCCESS;
    }

    // Set the m_SCIFNode value and connect to the device.
    if ( m_ISA != COI_ISA_x86_64 && !m_Connected )
    {
        m_SCIFNode = node;
        result = ConnectToDaemon();
    }
    return result;
}

COIRESULT _COIEngine::GetCount(COI_ISA_TYPE        in_ISA,
                               uint32_t*           out_pNumEngines)
{
    _PthreadAutoLock_t lock( _COIEngine::GetClassLock());
    switch (in_ISA)
    {
        case COI_ISA_x86_64:
            *out_pNumEngines = _COIEngine::internal_scif_get_num_x86_64_engines();
            break;
        case COI_ISA_MIC:
            *out_pNumEngines = _COIEngine::internal_scif_get_num_mic_engines();
            break;
        case COI_ISA_KNF:
            *out_pNumEngines = _COIEngine::internal_scif_get_num_knf_engines();
            break;
        case COI_ISA_KNC:
            *out_pNumEngines = _COIEngine::internal_scif_get_num_knc_engines();
            break;
        default:
            return COI_DOES_NOT_EXIST;
    }
    return COI_SUCCESS;
}


COIRESULT
_COIEngine::GetHandle(
            COI_ISA_TYPE        in_ISA,
            uint32_t            in_EngineIndex,
            COIENGINE*          out_pEngineHandle)
{
    COIRESULT result = COI_ERROR;
    _COIEngine* engine = NULL;
    uint16_t base = (uint16_t)-1;
    _PthreadAutoLock_t lock( _COIEngine::GetClassLock());
    switch (in_ISA)
    {
    case COI_ISA_x86_64:
        if( in_EngineIndex >= internal_scif_get_num_x86_64_engines() )
        {
            return COI_OUT_OF_RANGE;
        }
        engine = &(g_engines_x86_64[in_EngineIndex]);
        base = 0;
        engine->m_ISA = COI_ISA_x86_64;
        break;
    case COI_ISA_MIC:
        if( in_EngineIndex >= internal_scif_get_num_mic_engines() )
        {
            return COI_OUT_OF_RANGE;
        }
        if( internal_scif_get_num_knf_engines() )
        {
            engine = &(g_engines_KNF[in_EngineIndex]);
            base = COI_MAX_ISA_x86_64_DEVICES;
            engine->m_ISA = COI_ISA_KNF;
        }
        else
        {
            engine = &(g_engines_KNC[in_EngineIndex]);
            base = COI_MAX_ISA_x86_64_DEVICES;
            engine->m_ISA = COI_ISA_KNC;
        }
        break;
    case COI_ISA_KNF:
        if( in_EngineIndex >= internal_scif_get_num_knf_engines() )
        {
            return COI_OUT_OF_RANGE;
        }
        engine = &(g_engines_KNF[in_EngineIndex]);
        base = COI_MAX_ISA_x86_64_DEVICES;
        engine->m_ISA = COI_ISA_KNF;
        break;
    case COI_ISA_KNC:
        if( in_EngineIndex >= internal_scif_get_num_knc_engines() )
        {
            return COI_OUT_OF_RANGE;
        }
        engine = &(g_engines_KNC[in_EngineIndex]);
        base = COI_MAX_ISA_x86_64_DEVICES; //KNC devices will also be node 1+, they won't be node 5+
        engine->m_ISA = COI_ISA_KNC;
        break;
    default:
        return COI_DOES_NOT_EXIST;
    }

    // We set a pointer to the handle and set the m_ISA.
    // We need to actually set the m_SCIFnode and connect to that device.
    result = engine->SetScifNode(base + in_EngineIndex);
    if (result == COI_SUCCESS)
    {
        *out_pEngineHandle = (COIENGINE)engine;
    }

    return result;
}

COIRESULT _COIEngine::GetInfo(COI_ENGINE_INFO_BASE* out_pInfo)
{
    _PthreadAutoLock_t lock( GetLock() );

    // It's theoretically possible that they guessed the correct
    // address value that corresponds to a valid COIENGINE and then
    // called COIEngineGetInfo with that handle.
    if( !m_Connected )
    {
        COILOG_FUNC_RETURN_RESULT(COI_NOT_INITIALIZED);
    }

    COIDaemonMessage_t  message;
    COIRESULT result = COI_ERROR;

    // Send a request to the engine and have it send back information about
    // the resources on the engine.
    COIDaemonMessage_t::ENGINE_INFO_REQUEST_T* request;
    message.SetPayload(request);

    // Send the message to the daemon requesting engine info
    result = m_EngComm.SendMessageAndReceiveResponseAtomic(message, message);
    if (COI_SUCCESS != result )
    {
        // If the daemon died, reconnect and resend the message
        if( result == COI_PROCESS_DIED )
        {
            const char *coi_port_env = getenv("COI_DAEMON_PORT");
	    
            if (coi_port_env == 0 || coi_port_env[0] == '\0')
            {
                if( m_EngComm.Connect( GetScifNode(), COI_DAEMON_PORT, true) != -1 )
                {
                    result = m_EngComm.SendMessageAndReceiveResponseAtomic(message, message);
                }
	        }
            else
            {
                uint16_t coi_port = 0;
                stringstream ss(coi_port_env);
                ss >> coi_port;
                //ss returns null if th efirst term of coi_port isn't a number
                //ss ignores everything that isn't a # past the first term.
                //ss returns null if the number is > 65535
                //ss reroutes negatives to be added to 65535
                //don't allow the use of port 0
                if (!ss || coi_port < 1)
                {
                    COILOG_ERROR("$COI_DAEMON_PORT contains invalid port");
              	    result = COI_BAD_PORT;
                }
                else
                {
                    if( m_EngComm.Connect( GetScifNode(), coi_port, true) != -1 )
                    {
                        result = m_EngComm.SendMessageAndReceiveResponseAtomic(message, message);
                    }
                }
            }
        }
        COILOG_FUNC_RETURN_IF_ERROR(result);
    }

    if (COIDaemonMessage_t::ENGINE_INFO_RESULT != message.opcode())
    {
        out_pInfo = NULL;
        COILOG_FUNC_RETURN_ERROR(COI_ERROR);
    }

    COIDaemonMessage_t::ENGINE_INFO_RESULT_T* res;
    res = message.GetPayload();
    if (COI_SUCCESS != res->result)
    {
        result = (COIRESULT)res->result;
        COILOG_FUNC_RETURN_ERROR(result);
    }

    // TODO:  Again, in the future, it is possible that the sizeof the incoming
    // out_pEngineInfo is smaller than the COI_ENGINE_INFO when this code is
    // compiled.  therefore, only copy over the parts that will fit instead of
    // blindly copying it all over like this.
    *out_pInfo = res->engine_info;

    COILOG_FUNC_RETURN_RESULT(result);

}

COI_ISA_TYPE _COIEngine::GetIsaType()
{
    return m_ISA;
}

// Get the ELF header machine type that is compatible with the device
// that corresponds to this engine
Elf64_Ehdr_Machine::Elf64_Ehdr_Machine _COIEngine::GetElfMachineType()
{
    Elf64_Ehdr_Machine::Elf64_Ehdr_Machine type;
    switch(GetIsaType())
    {
    case COI_ISA_x86_64:
        type = Elf64_Ehdr_Machine::EM_X86_64;
        break;
    case COI_ISA_KNF:
        type = Elf64_Ehdr_Machine::EM_L1OM;
        break;
    case COI_ISA_KNC:
        type = Elf64_Ehdr_Machine::EM_K1OM;
        break;
    default:
        // an invalid machine number
        type = (Elf64_Ehdr_Machine::Elf64_Ehdr_Machine)0;
    }
    return type;
}

_COIEngine* _COIEngine::Get(COIENGINE engine)
{
    // Given a COIENGINE handle "engine", look for this
    // in our tables of handles "g_engines_*".
    _COIEngine* ptr = NULL;
    const _COIEngine* lookfor = (_COIEngine*)engine;
    for(size_t i = 0; i < COI_MAX_ISA_x86_64_DEVICES; i++ )
    {
        ptr = &g_engines_x86_64[i];
        if( lookfor == ptr )
        {
            return ptr;
        }
    }
    for(size_t i = 0; i < COI_MAX_ISA_KNC_DEVICES; i++ )
    {
        ptr = &g_engines_KNC[i];
        if( lookfor == ptr )
        {
            return ptr;
        }
    }
    for(size_t i = 0; i < COI_MAX_ISA_KNF_DEVICES; i++ )
    {
        ptr = &g_engines_KNF[i];
        if( lookfor == ptr )
        {
            return ptr;
        }
    }
    return NULL;
}

void _COIEngine::CleanUpHandler()
{
	_COIProcess* process;
    _PthreadAutoLock_t _l(m_cleanup_handler_lock);
	while(!coiProcessList.empty()){
		process = coiProcessList.front();
		process->CleanHost();
		coiProcessList.pop_front();
	}
}

COIRESULT _COIEngine::DeleteFromProcessList(_COIProcess* process)
{
	_PthreadAutoLock_t _l(m_cleanup_handler_lock);
	COIRESULT result = COI_ERROR;
	if(!coiProcessList.empty()) {
		coiProcessList.remove(process);
		result = COI_SUCCESS;
	}
	return result;
	
}

COIRESULT _COIEngine::AddToProcessList(_COIProcess* process)
{
	_PthreadAutoLock_t _l(m_cleanup_handler_lock);
    if(!m_handler_setup)
    {
        struct sigaction new_action, old_action;
        new_action.sa_handler = EngineCleanUpHandler;
        sigemptyset (&new_action.sa_mask);
        new_action.sa_flags = 0;
        sigaction(SIGINT, NULL, &old_action);
        if(old_action.sa_handler != SIG_IGN)
        sigaction(SIGINT, &new_action, NULL);
        m_handler_setup=true;
    }
	COIRESULT result = COI_ERROR;
    coiProcessList.push_back(process);
	result = COI_SUCCESS;
	return result;
	
}

uint32_t
_COIEngine::internal_scif_get_num_x86_64_engines()
{
    return 1;
}

uint32_t
_COIEngine::internal_scif_get_num_mic_engines()
{

    uint16_t self = 0;
    uint16_t online_nodes[COI_MAX_ISA_MIC_DEVICES] = {0};

    // This next variable really represents
    // "maximum index of scif node that I know about" + 1.
    // The current scif_get_nodeIDs API doesn't provide a lot of
    // info about offline nodes, but you can deduce that if
    // node N is online, then nodes 0 - (N-1) must exist.
    // There is currently some discussion about changing how
    // scif_get_nodeIDs works, so 
    // TODO - Modify this to be able to be even more dynamic, maybe.
    // RTC[33517]
    static int total_scif_nodes = 0;
    static int total_mic_devices = -1;

    if( total_mic_devices < 0 )
    {
        int online_node_count = 0;
        if( total_scif_nodes <= 0 )
        {
            total_scif_nodes = scif_get_nodeIDs(online_nodes,
                                                COI_MAX_ISA_MIC_DEVICES,
                                                &self);
        }
        if( total_scif_nodes <= 0 )
        {
            COILOG_ERROR("scif_get_nodeIDs returned %d nodes available\n", total_scif_nodes);
            return 0;
        }
        
        online_node_count = total_scif_nodes;

        // Node1 may have been offline but Node2 was available.
        // This is an error condition that we need to account for.
        for(int i = 0; i < online_node_count; i++)
        {
            total_scif_nodes = std::max(total_scif_nodes, online_nodes[i] + 1);
        }

        total_mic_devices = total_scif_nodes -
                            internal_scif_get_num_x86_64_engines();

        if( total_mic_devices > COI_MAX_ISA_MIC_DEVICES )
        {
            total_mic_devices = COI_MAX_ISA_MIC_DEVICES;
        }
        else if( total_mic_devices < 0 )
        {
            total_mic_devices = 0;
        }

    }

    return (uint32_t)total_mic_devices;
}

uint32_t
_COIEngine::internal_scif_get_num_knf_engines()
{
    static int total_knf_devices = -1;

    if (total_knf_devices >= 0)
    {
        return total_knf_devices;
    }
    else
    {
        // Use sysfs to check the family of the board.
        // This file contains either "Knights Ferry" (or "Knights Corner"?)
        FILE *f = fopen("/sys/class/mic/mic0/family","r");
        if (f == NULL) {
            // No boards are present: Not an error though.
            // C.f. CL 159344 and HSD[4115142].
            return 0;
        }

        char buf[32];
        if (fgets(buf,sizeof(buf),f) == NULL) {
            COILOG_ERROR("internal_scif_get_num_knf_engines(): "
                "read error in %s\n", __FILE__);
            fclose(f);
            return 0;
        }
        fclose(f);

        if (strncasecmp(buf,"Knights Ferry",strlen("Knights Ferry")) == 0)
        {
            total_knf_devices = internal_scif_get_num_mic_engines();
            return total_knf_devices;
        }
    }
    return 0;
}

// TODO - Whenever we support KNL you can't use the "if not KNF then it must be KNC"
//        shortcut.
uint32_t
_COIEngine::internal_scif_get_num_knc_engines()
{
    static int total_knc_devices = -1;
    if (total_knc_devices >= 0)
    {
        return total_knc_devices;
    }
    else
    {
        if (0 == internal_scif_get_num_knf_engines())
        {
            total_knc_devices = internal_scif_get_num_mic_engines();
            return total_knc_devices;
        }
    }
    return 0;
}


