/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */


#if SINK_LOG_ENABLED
#include <internal/_EnvHelper.h>
#endif
#include <internal/_Log.h>
#include <internal/_Process.h>
#include <internal/_Engine.h>
#include <internal/_Message.h>
#include <common/COIMacros_common.h>
#include <common/COIPerf_common.h>
#include <internal/_PthreadMutexAutoLock.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>


#include <unistd.h>

#include <stdarg.h>
#include <stdlib.h>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>

class COILogMessage_t : public OpcodeMessage_t
{
public:
    enum
    {
        INVALID = 0,
        SHUTDOWN,
        MESSAGE,
    };

    SUB_MESSAGE_TYPE(SHUTDOWN,
        // No fields
    );

    SUB_MESSAGE_TYPE(MESSAGE,
        // variable sized message
        char message[0];
    );
};

// Helpers
#define COILOG_LEVEL_STRINGIFY(COI_LOGLEVEL) # COI_LOGLEVEL

#define DEFAULT_LOG_FILE_PATTERN "%e.coilog"
#define MAX_NAME_SIZE       256
#define MAX_LOG_SIZE        2048

using namespace std;

    #if SINK_LOG_ENABLED
    using namespace EnvironmentHelper;
    #endif

// Set with default values
static bool g_log_enabled =
                    #ifdef DEBUG
                            true;
                    #else
                            false;
                    #endif

static uint16_t         g_verbosity   = COILOG_VERBOSITY_ALL;
static uint16_t         g_level       = COILOG_LEVEL_ERROR;
static volatile bool    g_initialized = false;
static volatile bool    g_initFailed  = false;
static __thread bool    t_in_log      = false;

#if SINK_LOG_ENABLED
static volatile bool    g_run_log     = true;
static _SCIFListener    g_listnr;
static _SCIFComm        g_comm;
#endif

// Keep a thread-specific tabstop around
static __thread uint16_t g_COILOG_TABSTOP = 0;

#if SINK_LOG_ENABLED
//static pthread_t          g_thread_handle = 0;
#endif
static pthread_mutex_t      g_log_mutex = PTHREAD_MUTEX_INITIALIZER;
static string               g_file_name;

//#ifdef _WIN32
//extern DWORD dwTlsIndex;
//#else


static const char* COILOG_LEVEL_STRINGS[] =
{    
    COILOG_LEVEL_STRINGIFY( COILOG_LEVEL_ERROR ),
    COILOG_LEVEL_STRINGIFY( COILOG_LEVEL_WARNING ),
    COILOG_LEVEL_STRINGIFY( COILOG_LEVEL_DETAIL ),
    COILOG_LEVEL_STRINGIFY( COILOG_LEVEL_TRACE )
};

static void COIWriteLogToFile(ostringstream& s)
{
    _PthreadAutoLock_t _l(g_log_mutex);

    // Create a new object and open for append
    fstream file(g_file_name.c_str(), fstream::out | fstream::app);
    file << s.str();
    file.close();
}

#if SINK_LOG_ENABLED
static void COIWriteToSource(ostringstream& s)
{
    COILogMessage_t message;

    // Logging is best effort so if the send queue is full it's ok to
    // drop the message.
    //
    _PthreadAutoLock_t _l(g_log_mutex);

    s<<'\0';

    COILogMessage_t::MESSAGE_T* msg;
    message.SetPayload(msg,s.str().length());

    strcpy(&msg->message[0],s.str().c_str());
    g_comm.Send(message);

}
#endif

static COIRESULT
COILogSourceInit()
{
    if (g_initialized)
    {
        return COI_SUCCESS;
    }
    // If it wasn't already initialized then need to make sure another
    // thread isn't in the middle of the initialization.
    //
    _PthreadAutoLock_t _l(g_log_mutex);
    if (g_initialized)
    {
        return COI_SUCCESS;
    }
    const char *enable_value = getenv("COI_ENABLE_LOGGING");
    if (enable_value)
    {
        g_log_enabled = strcmp(enable_value, "1") == 0 ;
    }
    const char *level_value = getenv("COI_LOG_LEVEL");

    if (level_value)
    {
        g_level = atoi(level_value) ;
    }

    if (!g_log_enabled)
    {
        // Stop things from coming in to this check
        g_initialized = true;
        return COI_ERROR;
    }

    // We use a log file pattern for the log file name. The pattern semantics
    // are similar to core files (man core(5)). We support a subset.
    //
    //   - Any non % character is take literally. Additionally, we expand
    //    %% to a % character
    //    %p to the pid (in decimal)
    //    %e to the executable
    //
    // By default, we use %e. However, this can be overriden with the environ-
    // ment variable COI_LOG_FILE_PATTERN.
    char          log_file[MAX_NAME_SIZE] = {0};
    const char    *log_file_pattern = getenv("COI_LOG_FILE_PATTERN");
    const char    *err_msg = NULL;
    int            err = 0;

    if (log_file_pattern == NULL) {
        log_file_pattern = DEFAULT_LOG_FILE_PATTERN;
    }

    char *s = log_file;
    const char *patt = log_file_pattern;
    while (*patt) {
        if (s - log_file == MAX_NAME_SIZE - 1) {
            err_msg = "log file name is too long";
            break;
        } else if (*patt != '%') {
            *s++ = *patt++;
        } else if (*(patt+1) == 'p') {
            s += snprintf(s, MAX_NAME_SIZE - (s - log_file) - 1,
                "%d", (int)getpid());
            patt += 2;
        } else if (*(patt+1) == 'e') {
            // expand %e to the executable name
            char exe[MAX_NAME_SIZE];
            ssize_t len;
            /* FIXME: how to get the name of the currently running executable
               on windows? */
            len = readlink("/proc/self/exe", exe, sizeof(exe));
            if (len == -1) {
                err = errno;
                err_msg = "readlink";
                snprintf(log_file, sizeof(log_file), "coilog");
                break;
            }
            exe[len] = 0;

            // proc/self/exe consists of a full path, take the filename only
            char *end = strrchr(exe,'/');
            if (end == NULL) {
                end = &exe[0];
            } else {
                end++; // we found the last /, skip it
            }

            size_t nleft = MAX_NAME_SIZE - (s - log_file);
            strncpy(s, end, nleft - 1);
            s[nleft - 1] = '\0';
            s += strlen(s);
            log_file[sizeof(log_file)-1] = 0;
            patt += 2;
        } else if (*(patt+1) == '%') {
            *s++ = '%';
            patt += 2;
        } else {
            // take last %, or the %x where x is an unsupported pattern
            *s++ = *patt++;
            err_msg = "malformed logfile name pattern, stray %";
            break;
        }
    }
    *s = 0;

    if (strlen(log_file) == 0) {
        strncpy(log_file, "coilog", sizeof(log_file) - 1);
        log_file[sizeof(log_file) - 1] = '\0';
        err_msg = "empty log file name defaulting to coilog";
    }

    g_file_name = log_file;
    FILE *f = fopen(log_file, "w");
    if (f == NULL) {
        // This is very possible, e.g. if root runs a test case, root owns the
        // log file. Then when mere mortals try and run it, the open will fail.
        fprintf(stderr, "COILog: when opening log file %s: %s\n",
            log_file, strerror(errno));
    }

    // Something went wrong building the file name, if we did report the
    // message there. Otherwise report the message to stderr.
    if (err_msg) {
        FILE *errf = f != NULL ? f : stderr;
        fprintf(errf, "error creating log file from pattern [%s]",
            log_file_pattern);
        if (err) {
            fprintf(errf, ": %s", strerror(err));
        }
        fprintf(errf, "\n");
    }

    // Close it for now...
    if (f) {
        fclose(f);
    }

    g_initialized = true;
    return COI_SUCCESS;
}

#if SINK_LOG_ENABLED
static COIRESULT
COILogSinkInit()
{
    COIRESULT result = COI_ERROR;
    int status = -1;
    uint32_t node = 0;
    uint32_t portNumber = 0;

    if (g_initialized)
    {
        return COI_SUCCESS;
    }

    // If it wasn't already initialized then need to make sure another
    // thread isn't in the middle of the initialization.
    //
    _PthreadAutoLock_t _l(g_log_mutex);
    if (g_initialized)
    {
        return COI_SUCCESS;
    }
    if (g_initFailed)
    {
        return COI_ERROR;
    }

    result = EnvHelper::GetEnv_("COI_SCIF_SOURCE_NODE", node);
    if (result != COI_SUCCESS)
    {
        g_initFailed = true;
        return COI_ERROR;
    }

    result = EnvHelper::GetEnv_("COI_LOG_PORT", portNumber);
    if (result != COI_SUCCESS)
    {
        g_initFailed = true;
        return COI_ERROR;
    }
    unsetenv("COI_LOG_PORT");

    status = g_comm.Bind(0);
    if( status == -1 )
    {
        g_initFailed = true;
        return COI_ERROR;
    }
    status = g_comm.Connect(node, portNumber);
    if (status < 0)
    {
        g_initFailed = true;
        perror("sink_connect: error");
        return COI_ERROR;
    }
    g_initialized = true;
    return COI_SUCCESS;
}
#endif

COIRESULT COILogSinkShutdown()
{
#if SINK_LOG_ENABLED

    COILogMessage_t message;

    COILogSinkInit();

    _PthreadAutoLock_t _l(g_log_mutex);

    COILogMessage_t::SHUTDOWN_T* msg;

    message.SetPayload(msg);

    g_comm.Send(message);
#endif

    return COI_SUCCESS;
}

# if 0
// These functions are only meant for debugging, removing for code coverage,
// but leaving here in case they are needed later

void COILOG_ENABLE(bool enable)
{
    _PthreadAutoLock_t _l(g_log_mutex);
    g_log_enabled = enable;
}

bool COILOG_IS_ENABLED()
{
    return g_log_enabled;
}

void COILOG_SET_VERBOSITY(int verbosity)
{
    _PthreadAutoLock_t _l(g_log_mutex);
    g_verbosity = verbosity;
}

void COILOG_SET_LEVEL(int level)
{
    _PthreadAutoLock_t _l(g_log_mutex);
    g_level = level;
}
# endif

void COILog(
    const   uint16_t            in_LogLevel,
    const   int8_t              in_Indent,
    const   char*               in_File,
    const   uint64_t            in_Line,
    const   char*               in_Func,
    const   char*               in_Preface,
    const   char*               in_LogMessage, ... )
{
    // Possible to override enabled status in init, so try that first
    int err = errno;
    if (!g_initialized)
    {
        if (COI_MODE_SOURCE == g_COI_MODE)
        {
            if (COI_SUCCESS != COILogSourceInit())
            {
                errno = err;
                return;
            }
        }
#if SINK_LOG_ENABLED
        else
        {
            if (COI_SUCCESS != COILogSinkInit())
            {
                errno = err;
                return;
            }
        }
#endif
    }

    if (!g_log_enabled)
    {
        return;
    }

    //char                currentTimeString[MAX_NAME_SIZE]; time_t currentTime;
    //struct tm*          timeInfo;
    char                buffer[MAX_LOG_SIZE];
    va_list             args;
    string              temp;


    pthread_t           in_ThreadID;


    // This will only be true if this thread is already in a log call.  This
    // prevents a recursive call into the logger if something in the logging
    // code makes a call to something that tries to log.
    //
    if (t_in_log)
    {
        return;
    }
    t_in_log = true;

    // ERROR messages are always logged
    if (g_level < in_LogLevel)
    {
        t_in_log = false;
        return;
    }
    in_ThreadID = pthread_self();
    
    if (in_Indent < 0 && g_COILOG_TABSTOP >= abs(in_Indent))
    {
        g_COILOG_TABSTOP -= abs(in_Indent);
    }

    // Statement format is:
    // [Source/Sink][ThreadID][Timestamp][File:Line][Log Level][Function]: message

    // Log statement can have any number of args,
    // retrieve all args
    va_start(args, in_LogMessage);
    vsnprintf(buffer, MAX_LOG_SIZE, in_LogMessage, args);
    va_end(args);

    ostringstream logStream;

    // Tab out to current tabstop
    for (int i = 0; i < g_COILOG_TABSTOP; i++)
    {
        logStream << "  ";
    }

    if (g_verbosity >= COILOG_VERBOSITY_ALL)
    {
        if (COI_MODE_SOURCE == g_COI_MODE)
        {
            logStream << "[SOURCE]";
        }
        else
        {
            logStream << "[SINK:"<<getpid()<<"]";
        }

        // Thread ID
        logStream << "[" << std::hex << "0x" << int(in_ThreadID) << std::dec << "]";

        // Timestamp
#if 0
        time(&currentTime);
        timeInfo = localtime(&currentTime);
        strftime(currentTimeString, MAX_NAME_SIZE, "%x-%X", timeInfo);
        logStream << "[" << currentTimeString << "]";
#endif
        logStream << "[" << COIPerfGetCycleCounter() << "]";
    }


    if (g_verbosity >= COILOG_VERBOSITY_FILE)
    {
        // File and line
        // Chop in_File down to just relative file name, no directories
        temp = in_File;

        // +1 gets rid of / at beginning of string
        temp = temp.erase(0, temp.rfind('/') + 1);

        logStream << "[" << temp.c_str() << ":" << in_Line << "]";
    }

    if (g_verbosity >= COILOG_VERBOSITY_LEVEL)
    {
        // Log level
        logStream << "[" << COILOG_LEVEL_STRINGS[in_LogLevel] << "]";
    }

    if (g_verbosity >= COILOG_VERBOSITY_FUNC)
    {
        // Function name
        logStream << "[" << in_Func << "]";
    }

    logStream << ": " << in_Preface;

    // Any log message passed in
    logStream << " " << buffer << "\n";

    if (COI_MODE_SOURCE == g_COI_MODE)
    {
        COIWriteLogToFile(logStream);
    }
#if SINK_LOG_ENABLED
    else
    {
        COIWriteToSource(logStream);
    }
#endif

    if (in_Indent > 0)
    {
        g_COILOG_TABSTOP += in_Indent;
    }

    t_in_log = false;
    errno = err;
}


#if SINK_LOG_ENABLED
void* _COILog::COILogSourceThread()
{
    int status = -1;
    ostringstream s;
    COIRESULT result;

    status = m_listnr.WaitForConnect(m_comm);
    if (-1 == status)
    {
        return NULL;
    }
    m_listnr.Close();
    PT_ASSERT(pthread_mutex_lock(&m_log_mutex));
    while (m_run_log)
    {
        COILogMessage_t message;
        PT_ASSERT(pthread_mutex_unlock(&m_log_mutex));
        result = m_comm.Receive(message);
        if (COI_SUCCESS != result)
        {
            return NULL;
        }
        if (COILogMessage_t::SHUTDOWN == message.opcode())
        {
            return NULL;
        }

        // cast it to a message struct
        COILogMessage_t::MESSAGE_T* log = message.GetPayload();

        char *log_message = &log->message[0];

        // wow, way too many messages here
        s << log_message;

        COIWriteLogToFile(s);

        s.str("");
		PT_ASSERT(pthread_mutex_lock(&m_log_mutex));
    }
        PT_ASSERT(pthread_mutex_unlock(&m_log_mutex));
    return NULL;
}
#endif

COIRESULT _COILog::COILogSourceStart(uint16_t *logPort)
{
    UNREFERENCED_PARAM(logPort);
#if SINK_LOG_ENABLED
    int status = -1;

    status = m_listnr.BindAndListen(0, 1);
    if (-1 == status)
    {
        return COI_ERROR;
    }

    *logPort = m_listnr.GetPortNumber();
    if (pthread_create(&m_thread_handle, NULL, Threadproc, this) != 0) {
        return COI_ERROR;
    }
#endif

    return COI_SUCCESS;
}

COIRESULT _COILog::COILogSourceShutdown()
{

#if SINK_LOG_ENABLED
    _PthreadAutoLock_t _l(m_log_mutex);

    m_run_log = false;
    m_listnr.Close();
    m_comm.Disconnect();
    if (m_thread_handle)
    {
        if (pthread_join(m_thread_handle, NULL) != 0) {
            return COI_ERROR;
        }
    }
    m_thread_handle = 0;
    m_run_log = true;
#endif

    return COI_SUCCESS;

}

