/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>

#include <errno.h>
#include <iostream>

#include <signal.h>
#include <pthread.h>
#include <sys/time.h>
#include <map>
#include <scif.h>

#include <common/COITypes_common.h>
#include <common/COIResult_common.h>
#include <internal/_Engine.h>
#include <internal/_PthreadMutexAutoLock.h>
#include <internal/_Proxy.h>
#include <internal/_Log.h>
#include <internal/_Process.h>


#include "uproxy.h"

using std::map;




// A synchronous version of scif_accept that can be canceled via,
// pThread_cancel. Technically, scif_accept calls into ioctl, there
// is no guarantee this will wake to a signal. Note, we allow a
// signal to interrupt poll for this case (since the threading library
// will likely cancel it with a signal).
static int scif_accept_cancelable(scif_epd_t ep,
    struct scif_portID *peer, scif_epd_t *conn, int timeout_ms)
{
    struct timeval t0;
    gettimeofday(&t0, NULL);

    struct scif_pollepd parr;
    parr.epd = ep;
    parr.events = SCIF_POLLIN;
    parr.revents = 0;
	
    int n = scif_poll(&parr, 1, timeout_ms);
    if (n == 0) {
        errno = ETIMEDOUT;
        return -1;
    } else if (n == -1) {
        return -1;
    } else if (parr.revents & SCIF_POLLIN) {
        int e = scif_accept(ep, peer, conn, SCIF_ACCEPT_SYNC);
        if (e == -1) {
            return -1;
        }
        return 0;
    } else if ((parr.revents & SCIF_POLLNVAL) || (parr.revents & SCIF_POLLERR)) {
        errno = EINVAL;
        return -1;
    } else {
        return -1;
    }
}

class COIProxy
{
public:
    COIProxy(scif_epd_t e) :
        endpoint(e), connection(SCIF_OPEN_FAILED), finish(false) {
    };

    void*
    WorkerThread(void)
    {
        struct scif_portID    peer;

        COILOG_INFO("waiting for uProxy connection");
        if (scif_accept_cancelable(endpoint,&peer,&connection,-1) == -1)
        {
            COILOG_ERROR("scif_accept failed errno %d", errno);
            finish = true;
        }
        else
        {
            COILOG_INFO("got uProxy connection == %d",connection);
        }

        while (!finish) {
            struct scif_pollepd poll_struct;

            poll_struct.epd = connection;
            poll_struct.events = SCIF_POLLIN;

            if (0 >= scif_poll(&poll_struct,1,100)) {
                // interrupted or something.
                continue;
            }
            proxy_call_args args;

            // RECV the header that tells us how many bytes will get sent
            int status = scif_recv(connection,&args,sizeof(args),RECV_BLOCK);
            if (-1 == status)
            {
                // Really? is this possible? scif_recv called with RECV_BLOCK
                // And EWOULDBLOCK is not listed as an errno value for
                // scif_recv.
                if (EWOULDBLOCK == errno)
                {
                    usleep(100);
                    continue;
                }
                COILOG_ERROR("scif_recv failed: %d",errno);
                break;
            }
            else if (0 == status)
            {
                break;
            }
            COILOG_INFO("got uProxy %d",args.c);
            if (uWRITE == args.c)
            {
                char  buf[PROXY_BUFFER_SIZE];
                char* ptr;
                int   nbytes;

                // RECV the actual output bytes that are getting proxied.
                // Note that we rely on the device not sending more than
                // PROXY_BUFFER_SIZE bytes at a time. Otherwise we'd need
                // to malloc and what not.
                status = scif_recv(connection,buf,args.nbytes,RECV_BLOCK);
                if (-1 == status)
                {
                    COILOG_ERROR("scif_recv failed: %d",errno);
                    break;
                }
                ptr = buf;
                // nbytes were received, let's print them out
                while (args.nbytes)
                {
                    nbytes = (int)write(args.fd,buf,args.nbytes);
                    
                    if (nbytes <= 0)
                    {
                        COILOG_ERROR("uProxy thread going away");
                        break;
                    }
                    args.nbytes -= nbytes;
                    ptr += nbytes;
                }

                // We wrote all the bytes, let's ACK that back to the other side.
                status = scif_send(connection,buf,1,SEND_BLOCK);
                
                if (-1 == status)
                {
                    COILOG_ERROR("scif_send failed: %d",errno);
                    break;
                }
            }
        }

        COILOG_INFO("exiting COIProxyThread");
        return (void*)0;
    }

    static void*
    WorkerThread(void* arg)
    {
        sigset_t ss;
        sigfillset(&ss);
        (void)pthread_sigmask(SIG_BLOCK,&ss,NULL);

        return ((COIProxy*)arg)->WorkerThread();
    }

    static COIProxy* Create( unsigned long pid, scif_epd_t endpoint )
    {
        COIProxy* proxy = new COIProxy(endpoint);

        if ( 0 > pthread_create(&proxy->thread, NULL, WorkerThread, proxy) )
        {
            COILOG_ERROR("pthread_create failed");
            delete proxy;
            proxy = NULL;
            return NULL;
        }

        _PthreadAutoLock_t l(m_lock);
        process_map[pid] = proxy;

        return proxy;
    }

    void End()
    {
        COILOG_INFO("ending thread");
        finish = true;
        if (connection != SCIF_OPEN_FAILED) {
            scif_close(connection);
        }
        scif_close(endpoint);

        // This will fail with ESRCH if the thread has already exited (the
        // normal case). An abnormal case where it is needed is, where
        // we must kill off the proxy before the sink-side has connected.
        (void) pthread_cancel(thread);
        PT_ASSERT(pthread_join(thread,NULL));
    }

    static COIRESULT Destroy( unsigned long pid )
    {
        COILOG_INFO("Destroying COIProxy for %ld",pid);
        map<unsigned long,COIProxy*>::iterator it;

        _PthreadAutoLock_t l(m_lock);
        it = process_map.find(pid);

        if (it == process_map.end())
        {
            COILOG_INFO("Could not find %lx",pid);
            return COI_DOES_NOT_EXIST;
        }

        COIProxy* proxy = it->second;
        proxy->End();
        delete proxy;
        proxy = NULL;

        process_map.erase(it);

        COILOG_INFO("Done");
        return COI_SUCCESS;
    }

private:
    static map<unsigned long,COIProxy*> process_map;
    scif_epd_t      endpoint;
    scif_epd_t      connection;
    volatile bool   finish;
    static pthread_mutex_t              m_lock;
    pthread_t                           thread;
};


map<unsigned long,COIProxy*>    COIProxy::process_map;
pthread_mutex_t                 COIProxy::m_lock = PTHREAD_MUTEX_INITIALIZER;


#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

COIRESULT
COIProxyCreateConnection(
    const   unsigned long   pid,
            uint16_t&       out_scif_proxy_node,
            uint16_t&       out_scif_proxy_port)
{
    scif_epd_t endpoint;
    int port_number;

    COILOG_INFO("creating process proxy connection");
    endpoint = scif_open();
    if ( SCIF_OPEN_FAILED == endpoint )
    {
        COILOG_ERROR("scif_open failed");
        return COI_RESOURCE_EXHAUSTED;
    }

    port_number = scif_bind(endpoint, 0);
    if ( -1 == port_number )
    {
        COILOG_ERROR("scif_bind failed");
        scif_close(endpoint);
        return COI_RESOURCE_EXHAUSTED;
    }
    COILOG_INFO("Bound to port %d", port_number);

    if ( -1 == scif_listen(endpoint,1) )
    {
        COILOG_ERROR("scif_listen failed");
        scif_close(endpoint);
        return COI_RESOURCE_EXHAUSTED;
    }

    if ( !COIProxy::Create(pid,endpoint) )
    {
        scif_close(endpoint);
        return COI_ERROR;
    }

    out_scif_proxy_node = _COIEngine::GetLocalScifNode();
    out_scif_proxy_port = port_number;

    return COI_SUCCESS;
}

COIRESULT
COIProxyDestroyConnection(unsigned long pid)
{
    return COIProxy::Destroy(pid);
}

#ifdef __cplusplus
} /* extern "C" */
#endif
